from __future__ import division, print_function
import numpy as np
import multiprocessing
from tools import _pickle_method, _unpickle_method
try:
    import copy_reg
except:
    import copyreg as copy_reg
import types

copy_reg.pickle(types.MethodType, _pickle_method, _unpickle_method)

class GeneticAlgorithm(object):
	"""Optimization algorithm for the EZ-Climate model. 

	Parameters
	----------
	pop_amount : int
		number of individuals in the population
	num_feature : int 
		number of elements in each individual, i.e. number of nodes in tree-model
	num_generations : int 
		number of generations of the populations to be evaluated
	bound : float
		upper bound of mitigation in each node
	cx_prob : float
		 probability of mating
	mut_prob : float
		probability of mutation.
	utility : `Utility` object
		object of utility class
	fixed_values : ndarray, optional
		nodes to keep fixed
	fixed_indicies : ndarray, optional
		indicies of nodes to keep fixed
	print_progress : bool, optional
		if the progress of the evolution should be printed

	Attributes
	----------
	pop_amount : int
		number of individuals in the population
	num_feature : int 
		number of elements in each individual, i.e. number of nodes in tree-model
	num_generations : int 
		number of generations of the populations to be evaluated
	bound : float
		upper bound of mitigation in each node
	cx_prob : float
		 probability of mating
	mut_prob : float
		probability of mutation.
	u : `Utility` object
		object of utility class
	fixed_values : ndarray, optional
		nodes to keep fixed
	fixed_indicies : ndarray, optional
		indicies of nodes to keep fixed
	print_progress : bool, optional
		if the progress of the evolution should be printed

	"""
	def __init__(self, pop_amount, num_generations, cx_prob, mut_prob, bound, num_feature, utility,
				 fixed_values=None, fixed_indicies=None, print_progress=False):
		self.num_feature = num_feature
		self.pop_amount = pop_amount
		self.num_gen = num_generations
		self.cx_prob = cx_prob
		self.mut_prob = mut_prob
		self.u = utility
		self.bound = bound
		self.fixed_values = fixed_values
		self.fixed_indicies = fixed_indicies
		self.print_progress = print_progress

	def _generate_population(self, size):
		"""Return 1D-array of random values in the given bound as the initial population."""
		pop = np.random.random([size, self.num_feature])*self.bound
		if self.fixed_values is not None:
			for ind in pop:
				ind[self.fixed_indicies] = self.fixed_values
		return pop

	def _evaluate(self, indvidual):
		"""Returns the utility of given individual."""
		return self.u.utility(indvidual)

	def _select(self, pop, rate):
		"""Returns a 1D-array of selected individuals.
	    
	    Parameters
	    ----------
	    pop : ndarray 
	    	population given by 2D-array with shape ('pop_amount', 'num_feature')
	    rate : float 
	    	the probability of an individual being selected
		    
	    Returns
	    -------
	    ndarray 
	    	selected individuals

		"""
		index = np.random.choice(self.pop_amount, int(rate*self.pop_amount), replace=False)
		return pop[index,:]

	def _random_index(self, individuals, size):
		"""Generate a random index of individuals of size 'size'.

		Parameters
		----------
		individuals : ndarray or list
			2D-array of individuals
		size : int 
			number of indices to generate
		
		Returns
		-------
		ndarray 
			1D-array of indices

		"""
		inds_size = len(individuals)
		return np.random.choice(inds_size, size)

	def _selection_tournament(self, pop, k, tournsize, fitness):
	    """Select `k` individuals from the input `individuals` using `k`
	    tournaments of `tournsize` individuals.
	    
	    Parameters
	    ----------
	    individuals : ndarray or list
	    	2D-array of individuals to select from
	    k : int
	    	 number of individuals to select
	    tournsize : int
	    	number of individuals participating in each tournament
	   
	   	Returns
	   	-------
	   	ndarray s
	   		selected individuals
	    
	    """
	    chosen = []
	    for i in xrange(k):
	        index = self._random_index(pop, tournsize)
	        aspirants = pop[index]
	        aspirants_fitness = fitness[index]
	        chosen_index = np.where(aspirants_fitness == np.max(aspirants_fitness))[0]
	        if len(chosen_index) != 0:
	        	chosen_index = chosen_index[0]
	        chosen.append(aspirants[chosen_index])
	    return np.array(chosen)

	def _two_point_cross_over(self, pop):
		"""Performs a two-point cross-over of the population.
	    
	    Parameters
	    ----------
		pop : ndarray 
			population given by 2D-array with shape ('pop_amount', 'num_feature')

		"""
		child_group1 = pop[::2]
		child_group2 = pop[1::2]
		for child1, child2 in zip(child_group1, child_group2):
			if np.random.random() <= self.cx_prob:
				cxpoint1 = np.random.randint(1, self.num_feature)
				cxpoint2 = np.random.randint(1, self.num_feature - 1)
				if cxpoint2 >= cxpoint1:
					cxpoint2 += 1
				else: # Swap the two cx points
					cxpoint1, cxpoint2 = cxpoint2, cxpoint1
				child1[cxpoint1:cxpoint2], child2[cxpoint1:cxpoint2] \
				= child2[cxpoint1:cxpoint2].copy(), child1[cxpoint1:cxpoint2].copy()
				if self.fixed_values is not None:
					child1[self.fixed_indicies] = self.fixed_values
					child2[self.fixed_indicies] = self.fixed_values
	
	def _uniform_cross_over(self, pop, ind_prob):
		"""Performs a uniform cross-over of the population.
	    
	    Parameters
	    ----------
	    pop : ndarray
	    	population given by 2D-array with shape ('pop_amount', 'num_feature')
	    ind_prob : float
	    	probability of feature cross-over
	    
		"""
		child_group1 = pop[::2]
		child_group2 = pop[1::2]
		for child1, child2 in zip(child_group1, child_group2):
			size = min(len(child1), len(child2))
			for i in range(size):
				if np.random.random() < ind_prob:
					child1[i], child2[i] = child2[i], child1[i]

	def _mutate(self, pop, ind_prob, scale=2.0):
		"""Mutates individual's elements. The individual has a probability of `mut_prob` of 
		beeing selected and every element in this individual has a probability `ind_prob` of beeing 
		mutated. The mutated value is a random number.

		Parameters
		----------
		pop : ndarray
			population given by 2D-array with shape ('pop_amount', 'num_feature')
	    ind_prob : float 
	    	probability of feature mutation 
	    scale : float 
	    	scaling constant of the random generated number for mutation

		"""
		pop_tmp = np.copy(pop)
		mutate_index = np.random.choice(self.pop_amount, int(self.mut_prob * self.pop_amount), replace=False)
		for i in mutate_index:
			feature_index = np.random.choice(self.num_feature, int(ind_prob * self.num_feature), replace=False)
			for j in feature_index:
				if self.fixed_indicies is not None and j in self.fixed_indicies:
					continue
				else:
					pop[i][j] = max(0.0, pop[i][j]+(np.random.random()-0.5)*scale)
	
	def _uniform_mutation(self, pop, ind_prob, scale=2.0):
		"""Mutates individual's elements. The individual has a probability of `mut_prob` of
		beeing selected and every element in this individual has a probability `ind_prob` of beeing 
		mutated. The mutated value is the current value plus a scaled uniform [-0.5,0.5] random value.

		Parameters
		----------
		pop : ndarray
			population given by 2D-array with shape ('pop_amount', 'num_feature')
	    ind_prob : float 
	    	probability of feature mutation 
	    scale : float 
	    	scaling constant of the random generated number for mutation

	    """ 
		pop_len = len(pop)
		mutate_index = np.random.choice(pop_len, int(self.mut_prob * pop_len), replace=False)
		for i in mutate_index:
			prob = np.random.random(self.num_feature) 
			inc = (np.random.random(self.num_feature) - 0.5)*scale
			pop[i] += (prob > (1.0-ind_prob)).astype(int)*inc
			pop[i] = np.maximum(1e-5, pop[i])
			if self.fixed_values is not None:
				pop[i][self.fixed_indicies] = self.fixed_values

	def _show_evolution(self, fits, pop):
		"""Print statistics of the evolution of the population."""
		length = len(pop)
		mean = fits.mean()
		std = fits.std()
		min_val = fits.min()
		max_val = fits.max()
		print (" Min {} \n Max {} \n Avg {}".format(min_val, max_val, mean))
		print (" Std {} \n Population Size {}".format(std, length))
		print (" Best Individual: ", pop[np.argmax(fits)])

	def _survive(self, pop_tmp, fitness_tmp):
		"""The 80 percent of the individuals with best fitness survives to
		the next generation.

		Parameters
		----------
		pop_tmp : ndarray
			population
		fitness_tmp : ndarray
			fitness values of `pop_temp`

		Returns
		-------
		ndarray 
			individuals that survived

		"""
		index_fits  = np.argsort(fitness_tmp)[::-1]
		fitness = fitness_tmp[index_fits]
		pop = pop_tmp[index_fits]
		num_survive = int(0.8*self.pop_amount) 
		survive_pop = np.copy(pop[:num_survive])
		survive_fitness = np.copy(fitness[:num_survive])
		return np.copy(survive_pop), np.copy(survive_fitness)

	def run(self):
		"""Start the evolution process.
		
		The evolution steps are:
			1. Select the individuals to perform cross-over and mutation.
			2. Cross over among the selected candidate.
			3. Mutate result as offspring.
			4. Combine the result of offspring and parent together. And selected the top 
			   80 percent of original population amount.
			5. Random Generate 20 percent of original population amount new individuals 
			   and combine the above new population.

		Returns
		-------
		tuple
			final population and the fitness for the final population

		Note
		----
		Uses the :mod:`~multiprocessing` package.

		"""
		print("----------------Genetic Evolution Starting----------------")
		pop = self._generate_population(self.pop_amount)
		pool = multiprocessing.Pool(processes=multiprocessing.cpu_count())
		fitness = pool.map(self._evaluate, pop) # how do we know pop[i] belongs to fitness[i]?
		fitness = np.array([val[0] for val in fitness])
		u_hist = np.zeros(self.num_gen)
		for g in range(0, self.num_gen):
			print ("-- Generation {} --".format(g+1))
			pop_select = self._select(np.copy(pop), rate=1)
			
			self._uniform_cross_over(pop_select, 0.50)
			self._uniform_mutation(pop_select, 0.25, np.exp(-float(g)/self.num_gen)**2)
			#self._mutate(pop_select, 0.05)
			
			fitness_select = pool.map(self._evaluate, pop_select)
			fitness_select = np.array([val[0] for val in fitness_select])
			
			pop_tmp = np.append(pop, pop_select, axis=0)
			fitness_tmp = np.append(fitness, fitness_select, axis=0)

			pop_survive, fitness_survive = self._survive(pop_tmp, fitness_tmp)

			pop_new = self._generate_population(self.pop_amount - len(pop_survive))
			fitness_new = pool.map(self._evaluate, pop_new)
			fitness_new = np.array([val[0] for val in fitness_new])

			pop = np.append(pop_survive, pop_new, axis=0)
			fitness = np.append(fitness_survive, fitness_new, axis=0)
			if self.print_progress:
				self._show_evolution(fitness, pop)
			u_hist[g] = fitness[0]

		fitness = pool.map(self._evaluate, pop)
		fitness = np.array([val[0] for val in fitness])
		return pop, fitness


class GradientSearch(object) :
	"""Gradient search optimization algorithm for the EZ-Climate model.

	Parameters
	----------
	utility : `Utility` object
		object of utility class
	learning_rate : float
		starting learning rate of gradient descent
	var_nums : int
		number of elements in array to optimize
	accuracy : float
		stop value for the gradient descent
	fixed_values : ndarray, optional
		nodes to keep fixed
	fixed_indicies : ndarray, optional
		indicies of nodes to keep fixed
	print_progress : bool, optional
		if the progress of the evolution should be printed
	scale_alpha : ndarray, optional
		array to scale the learning rate

	Attributes
	----------
	utility : `Utility` object
		object of utility class
	learning_rate : float
		starting learning rate of gradient descent
	var_nums : int
		number of elements in array to optimize
	accuracy : float
		stop value for the gradient descent
	fixed_values : ndarray, optional
		nodes to keep fixed
	fixed_indicies : ndarray, optional
		indicies of nodes to keep fixed
	print_progress : bool, optional
		if the progress of the evolution should be printed
	scale_alpha : ndarray, optional
		array to scale the learning rate

	"""

	def __init__(self, utility, var_nums, accuracy=1e-06, iterations=100, fixed_values=None, 
		        fixed_indicies=None, print_progress=False, scale_alpha=None):
		self.u = utility
		self.var_nums = var_nums
		self.accuracy = accuracy
		self.iterations = iterations
		self.fixed_values  = fixed_values
		self.fixed_indicies = fixed_indicies
		self.print_progress = print_progress
		self.scale_alpha = scale_alpha
		if scale_alpha is None:
			self.scale_alpha = np.exp(np.linspace(0.0, 3.0, var_nums))

	def _partial_grad(self, i):
		"""Calculate the ith element of the gradient vector."""
		m_copy = self.m.copy()
		m_copy[i] = m_copy[i] - self.delta if (m_copy[i] - self.delta)>=0 else 0.0
		minus_utility = self.u.utility(m_copy)
		m_copy[i] += 2*self.delta
		plus_utility = self.u.utility(m_copy)
		grad = (plus_utility-minus_utility) / (2*self.delta)
		return grad, i

	def numerical_gradient(self, m, delta=1e-08, fixed_indicies=None):
		"""Calculate utility gradient numerically.

		Parameters
		----------
		m : ndarray or list
			array of mitigation
		delta : float, optional
			change in mitigation 
		fixed_indicies : ndarray or list, optional
			indicies of gradient that should not be calculated

		Returns
		-------
		ndarray
			gradient

		"""
		self.delta = delta
		self.m = m
		if fixed_indicies is None:
			fixed_indicies = []
		grad = np.zeros(len(m))
		if not isinstance(m, np.ndarray):
			self.m = np.array(m)
		pool = multiprocessing.Pool()
		indicies = np.delete(range(len(m)), fixed_indicies)
		res = pool.map(self._partial_grad, indicies)
		for g, i in res:
			grad[i] = g
		pool.close()
		pool.join()
		del self.m
		del self.delta
		return grad

	def _accelerate_scale(self, accelerator, prev_grad, grad):
		sign_vector = np.sign(prev_grad * grad)
		scale_vector = np.ones(self.var_nums) * ( 1 + 0.10)
		accelerator[sign_vector <= 0] = 1
		accelerator *= scale_vector
		return accelerator


	def gradient_descent(self, initial_point, return_last=False):
		"""Gradient descent algorithm. The `initial_point` is updated using the 
		Adam algorithm. Adam uses the history of the gradient to compute individual 
		step sizes for each element in the mitigation vector. The vector of step 
		sizes are calculated using estimates of the first and second moments of 
		the gradient.

		Parameters
		----------
		initial_point : ndarray
			initial guess of the mitigation
		return_last : bool, optional
			if True the function returns the last point, else the point 
				with highest utility

		Returns
		-------
		tuple
			(best point, best utility)
		
		"""
		num_decision_nodes = initial_point.shape[0]
		x_hist = np.zeros((self.iterations+1, num_decision_nodes))
		u_hist = np.zeros(self.iterations+1)
		u_hist[0] = self.u.utility(initial_point)
		x_hist[0] = initial_point
		
		beta1, beta2 = 0.90, 0.90
		eta = 0.0015
		eps = 1e-3
		m_t, v_t = 0, 0

		prev_grad = 0.0
		accelerator = np.ones(self.var_nums)

		for i in range(self.iterations):
			grad = self.numerical_gradient(x_hist[i], fixed_indicies=self.fixed_indicies)
			m_t = beta1*m_t + (1-beta1)*grad
			v_t = beta2*v_t + (1-beta2)*np.power(grad, 2) 
			m_hat = m_t / (1-beta1**(i+1))
			v_hat = v_t / (1-beta2**(i+1))
			if i != 0:
				accelerator = self._accelerate_scale(accelerator, prev_grad, grad)
			
			new_x = x_hist[i] + ((eta*m_hat)/(np.square(v_hat)+eps)) * accelerator
			new_x[new_x < 0] = 0.0

			if self.fixed_values is not None:
				new_x[self.fixed_indicies] = self.fixed_values

			x_hist[i+1] = new_x
			u_hist[i+1] = self.u.utility(new_x)[0]
			prev_grad = grad.copy()

			if self.print_progress:
				print("-- Iteration {} -- \n Current Utility: {}".format(i+1, u_hist[i+1]))
				print(new_x)

		if return_last:
			return x_hist[i+1], u_hist[i+1]
		best_index = np.argmax(u_hist)
		return x_hist[best_index], u_hist[best_index]

	def run(self, initial_point_list, topk=4):
		"""Initiate the gradient search algorithm. 

		Parameters
		----------
		initial_point_list : list
			list of initial points to select from
		topk : int, optional
			select and run gradient descent on the `topk` first points of 
			`initial_point_list`

		Returns
		-------
		tuple
			best mitigation point and the utility of the best mitigation point

		Raises
		------
		ValueError
			If `topk` is larger than the length of `initial_point_list`.

		Note
		----
		Uses the :mod:`~multiprocessing` package.

		"""
		print("----------------Gradient Search Starting----------------")
		
		if topk > len(initial_point_list):
			raise ValueError("topk {} > number of initial points {}".format(topk, len(initial_point_list)))

		candidate_points = initial_point_list[:topk]
		mitigations = []
		utilities = np.zeros(topk)
		for cp, count in zip(candidate_points, range(topk)):
			if not isinstance(cp, np.ndarray):
				cp = np.array(cp)
			print("Starting process {} of Gradient Descent".format(count+1))
			m, u  = self.gradient_descent(cp)
			mitigations.append(m)
			utilities[count] = u
		best_index = np.argmax(utilities)
		return mitigations[best_index], utilities[best_index]


class CoordinateDescent(object):
	"""Coordinate descent optimization algorithm for the EZ-Climate model.

	Parameters
	----------
	utility : `Utility` object
		object of utility class
	var_nums : int
		number of elements in array to optimize
	accuracy : float
		stop value for the utility increase 
	iterations : int 
		maximum number of iterations

	Attributes
	----------
	utility : `Utility` object
		object of utility class
	var_nums : int
		number of elements in array to optimize
	accuracy : float
		stop value for the utility increase
	iterations : int 
		maximum number of iterations

	"""
	def __init__(self, utility, var_nums, accuracy=1e-4, iterations=100):
		self.u = utility
		self.var_nums = var_nums
		self.accuracy = accuracy
		self.iterations = iterations
	
	def _min_func(self, x, m, i):
		m_copy = m.copy()
		m_copy[i] = x
		return -self.u.utility(m_copy)[0]

	def _minimize_node(self, node, m):
		from scipy.optimize import fmin
		return fmin(self._min_func, x0=m[node], args=(m, node), disp=False)

	def run(self, m):
		"""Run the coordinate descent iterations.

		Parameters
		----------
		m : initial point

		Returns
		-------
		tuple
			best mitigation point and the utility of the best mitigation point

		Note
		----
		Uses the :mod:`~scipy` package.

		"""
		num_decision_nodes = m.shape[0]
		x_hist = []
		u_hist = []
		nodes = range(self.var_nums)
		x_hist.append(m.copy())
		u_hist.append(self.u.utility(m)[0])
		print("----------------Coordinate Descent Starting----------------")
		print("Starting Utility: {}".format(u_hist[0]))
		for i in range(self.iterations):
			print("-- Iteration {} --".format(i+1))
			node_iteration = np.random.choice(nodes, replace=False, size=len(nodes))
			for node in node_iteration:
				m[node] = max(0.0, self._minimize_node(node, m))
			x_hist.append(m.copy())
			u_hist.append(self.u.utility(m)[0])
			print("Current Utility: {}".format(u_hist[i+1]))
			if np.abs(u_hist[i+1] - u_hist[i]) < self.accuracy:
				break
		return x_hist[-1], u_hist[-1]


from __future__ import division
import numpy as np


class Forcing(object):
	"""Radiative forcing for the EZ-Climate model. Determines the excess energy created 
	by GHGs in the atmosphere. 

	Attributes
	----------
	sink_start : float
		sinking constant
	forcing_start : float
		forcing start constant
	forcing_p1 : float
		forcing constant
	forcing_p2 : float
		forcing constant 
	forcing_p3 : float
		forcing constant
	absorbtion_p1 : float
		absorbtion constant
	absorbtion_p2 : float 
		absorbtion constant
	lsc_p1 : float
		class constant
	lsc_p2 : float
		class constant 

	"""
	sink_start = 35.596
	forcing_start = 4.926
	forcing_p1 = 0.13173
	forcing_p2 = 0.607773
	forcing_p3 = 315.3785
	absorbtion_p1 = 0.94835
	absorbtion_p2 = 0.741547
	lsc_p1 = 285.6268
	lsc_p2 = 0.88414

	@classmethod
	def forcing_and_ghg_at_node(cls, m, node, tree, bau, subinterval_len, returning="forcing"):
		"""Calculates the radiative forcing based on GHG evolution leading up to the 
		damage calculation in `node`.

		Parameters
		----------
		m : ndarray
			array of mitigations
		node : int
			node for which forcing is to be calculated
		tree : `TreeModel` object 
			tree structure used
		bau : `BusinessAsUsual` object
			business-as-usual scenario of emissions
		subinterval_len : float
			subinterval length
		returning : string, optional
			* "forcing": implies only the forcing is returned
			* "ghg": implies only the GHG level is returned
			* "both": implies both the forcing and GHG level is returned

        Returns
        -------
        tuple or float
        	if `returning` is 
        		* "forcing": only the forcing is returned
        		* "ghg": only the GHG level is returned
        		* "both": both the forcing and GHG level is returned 

		"""
		if node == 0:
			if returning == "forcing":
				return 0.0
			elif returning== "ghg":
				return bau.ghg_start
			else:
				return 0.0, bau.ghg_start

		period = tree.get_period(node)
		path = tree.get_path(node, period)

		period_lengths = tree.decision_times[1:period+1] - tree.decision_times[:period]
		increments = period_lengths/subinterval_len

		cum_sink = cls.sink_start
		cum_forcing = cls.forcing_start
		ghg_level = bau.ghg_start

		for p in range(0, period):
			start_emission = (1.0 - m[path[p]]) * bau.emission_by_decisions[p]
			if p < tree.num_periods-1: 
				end_emission = (1.0 - m[path[p]]) * bau.emission_by_decisions[p+1]
			else:
				end_emission = start_emission
			increment = int(increments[p])
			for i in range(0, increment):
				p_co2_emission = start_emission + i * (end_emission-start_emission) / increment
				p_co2 = 0.71 * p_co2_emission 
				p_c = p_co2 / 3.67 
				add_p_ppm = subinterval_len * p_c / 2.13
				lsc = cls.lsc_p1 + cls.lsc_p2 * cum_sink
				absorbtion = 0.5 * cls.absorbtion_p1 * np.sign(ghg_level-lsc) * np.abs(ghg_level-lsc)**cls.absorbtion_p2
				cum_sink += absorbtion
				cum_forcing += cls.forcing_p1*np.sign(ghg_level-cls.forcing_p3)*np.abs(ghg_level-cls.forcing_p3)**cls.forcing_p2
				ghg_level += add_p_ppm - absorbtion

		if returning == "forcing":
			return cum_forcing
		elif returning == "ghg":
			return ghg_level
		else:
			return cum_forcing, ghg_level
	
	@classmethod
	def forcing_at_node(cls, m, node, tree, bau, subinterval_len):
		"""Calculates the forcing based mitigation leading up to the 
		damage calculation in `node`.

		Parameters
		----------
		m : ndarray 
			array of mitigations in each node. 
		node : int 
			the node for which the forcing is being calculated.

		Returns
		-------
		float 
			forcing 

		"""

		return cls.forcing_and_ghg_at_node(m, node, tree, bau, subinterval_len, returning="forcing")

	@classmethod
	def ghg_level_at_node(cls, m, node, tree, bau, subinterval_len):
		"""Calculates the GHG level leading up to the damage calculation in `node`.

		Parameters
		----------
		m : ndarray 
			array of mitigations in each node. 
		node : int 
			the node for which the GHG level is being calculated.

		Returns
		-------
		float 
			GHG level at node

		"""
		return cls.forcing_and_ghg_at_node(m, node,tree, bau, subinterval_len, returning="ghg")
		



# -*- coding: utf-8 -*-
import numpy as np
import csv

###########
### I/O ###
###########

def find_path(file_name, directory="data", file_type=".csv"):
	import os
	cwd = os.getcwd()
	if not os.path.exists(directory):
		os.makedirs(directory)
	d = os.path.join(cwd, os.path.join(directory,file_name+file_type))
	return d

def create_file(file_name):
	import os
	d = find_path(file_name)
	if not os.path.isfile(d):
		open(d, 'w').close()
	return d

def file_exists(file_name):
	import os
	d = find_path(file_name)
	return os.path.isfile(d)

def load_csv(file_name, delimiter=';', comment=None):
	d = find_path(file_name)
	pass

def write_columns_csv(lst, file_name, header=[], index=None, start_char=None, delimiter=';', open_as='wb'):
	d = find_path(file_name)
	if index is not None:
		index.extend(lst)
		output_lst = zip(*index)
	else:
		output_lst = zip(*lst)

	with open(d, open_as) as f:
		writer = csv.writer(f, delimiter=delimiter)
		if start_char is not None:
			writer.writerow([start_char])
		if header:
			writer.writerow(header)
		for row in output_lst:
			writer.writerow(row)

def write_columns_to_existing(lst, file_name, header="", delimiter=';'):
	d = find_path(file_name)
	with open(d, 'r') as finput:
			reader = csv.reader(finput, delimiter=delimiter)
			all_lst = []
			row = next(reader)
			nested_list = isinstance(lst[0], list) or isinstance(lst[0], np.ndarray)
			if nested_list:
				lst = zip(*lst)
				row.extend(header)	
			else:
				row.append(header)
			all_lst.append(row)
			n = len(lst)
			i = 0
			for row in reader:
				if nested_list:
					row.extend(lst[i])
				else:
					row.append(lst[i])
				all_lst.append(row)
				i += 1
	with open(d, 'w') as foutput:
			writer = csv.writer(foutput, delimiter=delimiter)
			writer.writerows(all_lst)
			
def append_to_existing(lst, file_name, header="", index=None, delimiter=';', start_char=None):
	write_columns_csv(lst, file_name, header, index, start_char=start_char, delimiter=delimiter, open_as='a')

def import_csv(file_name, delimiter=';', header=True, indices=None, start_at=0, break_at='\n', ignore=""):
	d = find_path(file_name)
	input_lst = []
	indices_lst = []
	with open(d, 'r') as f:
		reader = csv.reader(f, delimiter=delimiter)
		for _ in range(0, start_at):
			next(reader)
		if header:
			header_row = next(reader)
		for row in reader:
			if row[0] == break_at:
				break
			if row[0] == ignore:
				continue
			if indices:
				input_lst.append(row[indices:])
				indices_lst.append(row[:indices])
			else:
				input_lst.append(row)
	if header and not indices :
		return header_row, np.array(input_lst, dtype="float64")
	elif header and indices:
		return header_row[indices:], indices_lst, np.array(input_lst, dtype="float64")
	return np.array(input_lst, dtype="float64")


##########
### MP ###
##########

def _pickle_method(method):
    func_name = method.im_func.__name__
    obj = method.im_self
    cls = method.im_class
    if func_name.startswith('__') and not func_name.endswith('__'): #deal with mangled names
        cls_name = cls.__name__.lstrip('_')
        func_name = '_' + cls_name + func_name
    return _unpickle_method, (func_name, obj, cls)

def _unpickle_method(func_name, obj, cls):
    for cls in cls.__mro__:
        try:
            func = cls.__dict__[func_name]
        except KeyError:
            pass
        else:
            break
    return func.__get__(obj, cls)




hard_dependencies = ("numpy", )
missing_dependencies = []

for dependency in hard_dependencies:
    try:
        __import__(dependency)
    except ImportError as e:
        missing_dependencies.append(dependency)

if missing_dependencies:
    raise ImportError("Missing required dependencies {0}".format(missing_dependencies))

from optimization import *
from analysis import *
from bau import *
from cost import *
from damage import *
from damage_simulation import *
from forcing import *
from storage_tree import *
from tree import *
from utility import *




from __future__ import division, print_function
import numpy as np
from abc import ABCMeta, abstractmethod
from damage_simulation import DamageSimulation
from forcing import Forcing

class Damage(object):
	"""Abstract damage class for the EZ-Climate model.

	Parameters
	----------
	tree : `TreeModel` object
		provides the tree structure used
	bau : `BusinessAsUsual` object
		business-as-usual scenario of emissions

	Attributes
	----------
	tree : `TreeModel` object
		provides the tree structure used
	bau : `BusinessAsUsual` object
		business-as-usual scenario of emissions

	"""
	__metaclass__ = ABCMeta
	def __init__(self, tree, bau):
		self.tree = tree
		self.bau = bau

	@abstractmethod
	def average_mitigation(self):
		"""The average_mitigation function should return a 1D array of the
		average mitigation for every node in the period.
		"""
		pass

	@abstractmethod
	def damage_function(self):
		"""The damage_function should return a 1D array of the damages for
		every node in the period.
		"""
		pass

class DLWDamage(Damage):
	"""Damage class for the EZ-Climate model. Provides the damages from emissions and mitigation outcomes.

	Parameters
	----------
	tree : `TreeModel` object
		provides the tree structure used
	bau : `BusinessAsUsual` object
		business-as-usual scenario of emissions
	cons_growth : float
		constant consumption growth rate
	ghg_levels : ndarray or list
		end GHG levels for each end scenario

	Attributes
	----------
	tree : `TreeModel` object
		provides the tree structure used
	bau : `BusinessAsUsual` object
		business-as-usual scenario of emissions
	cons_growth : float
		constant consumption growth rate
	ghg_levels : ndarray or list
		end GHG levels for each end scenario
	dnum : int 
		number of simulated damage paths
	d : ndarray
		simulated damages 
	cum_forcing : ndarray
		cumulative forcing interpolation coeffiecients, used to calculate forcing based mitigation 
	forcing : `Forcing` object
		class for calculating cumulative forcing and GHG levels
	damage_coefs : ndarray
		interpolation coefficients used to calculate damages

	"""

	def __init__(self, tree, bau, cons_growth, ghg_levels, subinterval_len):
		super(DLWDamage, self).__init__(tree, bau)
		self.ghg_levels = ghg_levels
		if isinstance(self.ghg_levels, list):
			self.ghg_levels = np.array(self.ghg_levels)
		self.cons_growth = cons_growth
		self.dnum = len(ghg_levels)
		self.subinterval_len = subinterval_len
		self.cum_forcings = None
		self.d = None
		self.emit_pct = None
		self.damage_coefs = None

	def _recombine_nodes(self):
		"""Creating damage coefficients for recombining tree. The state reached by an up-down move is
		separate from a down-up move because in general the two paths will lead to different degrees of 
		mitigation and therefore of GHG level. A 'recombining' tree is one in which the movement from 
		one state to the next through time is nonetheless such that an up move followed by a down move 
		leads to the same fragility. 
        """
		nperiods = self.tree.num_periods
		sum_class = np.zeros(nperiods, dtype=int)
		new_state = np.zeros([nperiods, self.tree.num_final_states], dtype=int)
		temp_prob = self.tree.final_states_prob.copy()

		for old_state in range(self.tree.num_final_states):
			temp = old_state
			n = nperiods-2
			d_class = 0
			while n >= 0:
				if temp >= 2**n:
					temp -= 2**n
					d_class += 1
				n -= 1
			sum_class[d_class] += 1
			new_state[d_class, sum_class[d_class]-1] = old_state
		
		sum_nodes = np.append(0, sum_class.cumsum())
		prob_sum = np.array([self.tree.final_states_prob[sum_nodes[i]:sum_nodes[i+1]].sum() for i in range(len(sum_nodes)-1)])
		for period in range(nperiods):
			for k in range(self.dnum):
				d_sum = np.zeros(nperiods)
				old_state = 0
				for d_class in range(nperiods):
					d_sum[d_class] = (self.tree.final_states_prob[old_state:old_state+sum_class[d_class]] \
						 			 * self.d[k, old_state:old_state+sum_class[d_class], period]).sum()	
					old_state += sum_class[d_class]
					self.tree.final_states_prob[new_state[d_class, 0:sum_class[d_class]]] = temp_prob[0]
				for d_class in range(nperiods):	
					self.d[k, new_state[d_class, 0:sum_class[d_class]], period] = d_sum[d_class] / prob_sum[d_class]

		self.tree.node_prob[-len(self.tree.final_states_prob):] = self.tree.final_states_prob
		for p in range(1,nperiods-1):
			nodes = self.tree.get_nodes_in_period(p)
			for node in range(nodes[0], nodes[1]+1):
				worst_end_state, best_end_state = self.tree.reachable_end_states(node, period=p)
				self.tree.node_prob[node] = self.tree.final_states_prob[worst_end_state:best_end_state+1].sum()

	def _damage_interpolation(self):
		"""Create the interpolation coeffiecients used in `damage_function`."""
		if self.d is None:
			print("Importing stored damage simulation")
			self.import_damages()

		self._recombine_nodes()
		if self.emit_pct is None:
			bau_emission = self.bau.ghg_end - self.bau.ghg_start
			self.emit_pct = 1.0 - (self.ghg_levels-self.bau.ghg_start) / bau_emission
		
		self.damage_coefs = np.zeros((self.tree.num_final_states, self.tree.num_periods, self.dnum-1, self.dnum))
		amat = np.ones((self.tree.num_periods, self.dnum, self.dnum))
		bmat = np.ones((self.tree.num_periods, self.dnum))

		self.damage_coefs[:, :, -1,  -1] = self.d[-1, :, :]
		self.damage_coefs[:, :, -1,  -2] = (self.d[-2, :, :] - self.d[-1, :, :]) / self.emit_pct[-2]
		amat[:, 0, 0] = 2.0 * self.emit_pct[-2]
		amat[:, 1:, 0] = self.emit_pct[:-1]**2
		amat[:, 1:, 1] = self.emit_pct[:-1]
		amat[:, 0, -1] = 0.0

		for state in range(0, self.tree.num_final_states):
			bmat[:, 0] = self.damage_coefs[state, :, -1,  -2] * self.emit_pct[-2]
			bmat[:, 1:] = self.d[:-1, state, :].T
			self.damage_coefs[state, :, 0] = np.linalg.solve(amat, bmat)

	def import_damages(self, file_name="simulated_damages"):
		"""Import saved simulated damages. File must be saved in 'data' directory
		inside current working directory. Save imported values in `d`. 

		Parameters
		----------
		file_name : str, optional
			name of file of saved simulated damages

		Raises
		------
		IOError
			If file does not exist.

		"""
		from tools import import_csv
		try:
			d = import_csv(file_name, ignore="#", header=False)
		except IOError as e:
			import sys
			print("Could not import simulated damages:\n\t{}".format(e))
			sys.exit(0)

		n = self.tree.num_final_states	
		self.d = np.array([d[n*i:n*(i+1)] for i in range(0, self.dnum)])

	def damage_simulation(self, draws, peak_temp=9.0, disaster_tail=12.0, tip_on=True, 
		temp_map=1, temp_dist_params=None, maxh=100.0, save_simulation=True):
		"""Initializion and simulation of damages, given by :mod:`ez_climate.DamageSimulation`.

		Parameters
		----------
		draws : int
			number of Monte Carlo draws
		peak_temp : float, optional 
			tipping point parameter 
	    disaster_tail : float, optional
	    	curvature of tipping point
	    tip_on : bool, optional
	    	flag that turns tipping points on or off
	    temp_map : int, optional
	    	mapping from GHG to temperature
	            * 0: implies Pindyck displace gamma
	            * 1: implies Wagner-Weitzman normal
	            * 2: implies Roe-Baker
	            * 3: implies user-defined normal 
	            * 4: implies user-defined gamma
	    temp_dist_params : ndarray or list, optional
	    	if temp_map is either 3 or 4, user needs to define the distribution parameters
	    maxh : float, optional
	    	time paramter from Pindyck which indicates the time it takes for temp to get half 
	            way to its max value for a given level of ghg
	    cons_growth : float, optional 
	    	yearly growth in consumption
	    save_simulation : bool, optional
	    	True if simulated values should be save, False otherwise
		
	    Returns
	    -------
	    ndarray
	    	simulated damages

		"""
		ds = DamageSimulation(tree=self.tree, ghg_levels=self.ghg_levels, peak_temp=peak_temp,
					disaster_tail=disaster_tail, tip_on=tip_on, temp_map=temp_map, 
					temp_dist_params=temp_dist_params, maxh=maxh, cons_growth=self.cons_growth)
		print("Starting damage simulation..")
		self.d = ds.simulate(draws, write_to_file = save_simulation)
		print("Done!")
		return self.d

	def _forcing_based_mitigation(self, forcing, period): 
		"""Calculation of mitigation based on forcing up to period. Interpolating between the forcing associated 
		with the constant degree of mitigation consistent with the damage simulation scenarios.
		"""
		p = period - 1
		if forcing > self.cum_forcings[p][1]:
			weight_on_sim2 = (self.cum_forcings[p][2] - forcing) / (self.cum_forcings[p][2] - self.cum_forcings[p][1])
			weight_on_sim3 = 0
		elif forcing > self.cum_forcings[p][0]:
			weight_on_sim2 = (forcing - self.cum_forcings[p][0]) / (self.cum_forcings[p][1] - self.cum_forcings[p][0])
			weight_on_sim3 = (self.cum_forcings[p][1] - forcing) / (self.cum_forcings[p][1] - self.cum_forcings[p][0])
		else:
			weight_on_sim2 = 0
			weight_on_sim3 = 1.0 + (self.cum_forcings[p][0] - forcing) / self.cum_forcings[p][0]
		
		return weight_on_sim2 * self.emit_pct[1] + weight_on_sim3*self.emit_pct[0]

	def _forcing_init(self):
		"""Initialize `Forcing` object and cum_forcings used in calculating the force mitigation up to a node.""" 
		if self.emit_pct is None:
			bau_emission = self.bau.ghg_end - self.bau.ghg_start
			self.emit_pct = 1.0 - (self.ghg_levels-self.bau.ghg_start) / bau_emission

		self.cum_forcings = np.zeros((self.tree.num_periods, self.dnum))
		mitigation = np.ones((self.dnum, self.tree.num_decision_nodes)) * self.emit_pct[:, np.newaxis]

		for i in range(0, self.dnum):
			for n in range(1, self.tree.num_periods+1):
				node = self.tree.get_node(n, 0)
				self.cum_forcings[n-1, i] = Forcing.forcing_at_node(mitigation[i], node, self.tree,
																	self.bau, self.subinterval_len)

	def average_mitigation_node(self, m, node, period=None):
		"""Calculate the average mitigation until node.

		Parameters
		----------
		m : ndarray or list
			array of mitigation
		node : int
			node for which average mitigation is to be calculated for
		period : int, optional
			the period the node is in
	
		Returns
		-------
		float
			average mitigation

		"""
		if period == 0:
			return 0
		if period is None:
			period = self.tree.get_period(node)
		state = self.tree.get_state(node, period)
		path = self.tree.get_path(node, period)
		new_m = m[path[:-1]]
	
		period_len = self.tree.decision_times[1:period+1] - self.tree.decision_times[:period]
		bau_emissions = self.bau.emission_by_decisions[:period]
		total_emission = np.dot(bau_emissions, period_len)
		ave_mitigation = np.dot(new_m, bau_emissions*period_len)
		return ave_mitigation / total_emission

	def average_mitigation(self, m, period):
		"""Calculate the average mitigation for all node in a period.

		m : ndarray or list
			array of mitigation
		period : int
			period to calculate average mitigation for
		
		Returns
		-------
		ndarray
			average mitigations 

		"""
		nodes = self.tree.get_num_nodes_period(period)
		ave_mitigation = np.zeros(nodes)
		for i in range(nodes):
			node = self.tree.get_node(period, i)
			ave_mitigation[i] = self.average_mitigation_node(m, node, period)
		return ave_mitigation

	def _ghg_level_node(self, m, node):
		return Forcing.ghg_level_at_node(m, node, self.tree, self.bau, self.subinterval_len)

	def ghg_level_period(self, m, period=None, nodes=None):
		"""Calculate the GHG levels corresponding to the given mitigation.
		Need to provide either `period` or `nodes`.

		Parameters
		----------
		m : ndarray or list
			array of mitigation
		period : int, optional
			what period to calculate GHG levels for
		nodes : ndarray or list, optional
			the nodes to calculate GHG levels for
		
		Returns
		-------
		ndarray
			GHG levels

		"""
		if nodes is None and period is not None:
			start_node, end_node = self.tree.get_nodes_in_period(period)
			if period >= self.tree.num_periods:
				add = end_node-start_node+1
				start_node += add
				end_node += add
			nodes = np.array(range(start_node, end_node+1))
		if period is None and nodes is None:
			raise ValueError("Need to give function either nodes or the period")

		ghg_level = np.zeros(len(nodes))
		for i in range(len(nodes)):
			ghg_level[i] = self._ghg_level_node(m, nodes[i])
		return ghg_level

	def ghg_level(self, m, periods=None):
		"""Calculate the GHG levels for more than one period.

		Parameters
		----------
		m : ndarray or list
			array of mitigation
		periods : int, optional
			number of periods to calculate GHG levels for
		
		Returns
		-------
		ndarray
			GHG levels 

		"""
		if periods is None:
			periods = self.tree.num_periods-1
		if periods >= self.tree.num_periods:
			ghg_level = np.zeros(self.tree.num_decision_nodes+self.tree.num_final_states)
		else:
			ghg_level = np.zeros(self.tree.num_decision_nodes)
		for period in range(periods+1):
			start_node, end_node = self.tree.get_nodes_in_period(period)
			if period >= self.tree.num_periods:
				add = end_node-start_node+1
				start_node += add
				end_node += add
			nodes = np.array(range(start_node, end_node+1))
			ghg_level[nodes] = self.ghg_level_period(m, nodes=nodes)
		return ghg_level

	def _damage_function_node(self, m, node):
		"""Calculate the damage at any given node, based on mitigation actions in `m`."""
		if self.damage_coefs is None:
			self._damage_interpolation()
		if self.cum_forcings is None:
			self._forcing_init()
		if node == 0:
			return 0.0

		period = self.tree.get_period(node)
		forcing, ghg_level = Forcing.forcing_and_ghg_at_node(m, node, self.tree, self.bau, self.subinterval_len, "both")
		force_mitigation = self._forcing_based_mitigation(forcing, period)
		ghg_extension = 1.0 / (1 + np.exp(0.05*(ghg_level-200)))

		worst_end_state, best_end_state = self.tree.reachable_end_states(node, period=period)
		probs = self.tree.final_states_prob[worst_end_state:best_end_state+1]

		if force_mitigation < self.emit_pct[1]:
			damage = (probs *(self.damage_coefs[worst_end_state:best_end_state+1, period-1, 1, 1] * force_mitigation \
					 + self.damage_coefs[worst_end_state:best_end_state+1, period-1, 1, 2])).sum()
		
		elif force_mitigation < self.emit_pct[0]:
			damage = (probs * (self.damage_coefs[worst_end_state:best_end_state+1, period-1, 0, 0] * force_mitigation**2 \
					  + self.damage_coefs[worst_end_state:best_end_state+1, period-1, 0, 1] * force_mitigation \
					  + self.damage_coefs[worst_end_state:best_end_state+1, period-1, 0, 2])).sum()
		else: 
			damage = 0.0
			i = 0
			for state in range(worst_end_state, best_end_state+1): 
				if self.d[0, state, period-1] > 1e-5:
					deriv = 2.0 * self.damage_coefs[state, period-1, 0, 0]*self.emit_pct[0] \
							+ self.damage_coefs[state, period-1, 0, 1]
					decay_scale = deriv / (self.d[0, state, period-1]*np.log(0.5))
					dist = force_mitigation - self.emit_pct[0] + np.log(self.d[0, state, period-1]) \
						   / (np.log(0.5) * decay_scale) 
					damage += probs[i] * (0.5**(decay_scale*dist) * np.exp(-np.square(force_mitigation-self.emit_pct[0])/60.0))
				i += 1

		return (damage / probs.sum()) + ghg_extension

	def damage_function(self, m, period):
		"""Calculate the damage for every node in a period, based on mitigation actions `m`.

		Parameters
		----------
		m : ndarray or list
			array of mitigation
		period : int
			period to calculate damages for
		
		Returns
		-------
		ndarray
			damages

		"""
		nodes = self.tree.get_num_nodes_period(period)
		damages = np.zeros(nodes)
		for i in range(nodes):
			node = self.tree.get_node(period, i)
			damages[i] = self._damage_function_node(m, node)
		return damages 





import numpy as np
from abc import ABCMeta, abstractmethod

class BusinessAsUsual(object):
    """Abstract BAU class for the EZ-Climate model.

    Parameters
    ----------
    ghg_start : float
        today's GHG-level
    ghg_end : float
        GHG-level in the last period
    
    Attributes
    ----------
    ghg_start : float
        today's GHG-level
    ghg_end : float
        GHG-level in the last period
    emission_by_decisions : ndarray
        emissions at decision time periods??
    emission_per_period : ndarray
        total emission at decision time period??
    emission_to_ghg : ndarray
        GHG levels in decision time period??
    emission_to_bau : float
        constant for converting GHG to emission??

    """
    __metaclass__ = ABCMeta
    def __init__(self, ghg_start, ghg_end):
        self.ghg_start = ghg_start
        self.ghg_end = ghg_end
        self.emission_by_decisions = None
        self.emission_per_period = None
        self.emission_to_ghg = None
        self.emission_to_bau = None
        self.bau_path = None

    @abstractmethod
    def emission_by_time(self):
        pass


class DLWBusinessAsUsual(BusinessAsUsual):
    """Business-as-usual scenario of emissions. Emissions growth is assumed to slow down 
    exogenously - these assumptions represent an attempt to model emissions growth in a 
    business-as-usual scenario that is in the absence of incentives.

    Parameters
    ----------
    ghg_start : float
        today's GHG-level
    ghg_end : float
        GHG-level in the last period
    emit_time : ndarray or list
        time, in years, from now when emissions occurs
    emit_level : ndarray or list
        emission levels in future times `emit_time`

    Attributes
    ----------
    ghg_start : float
        today's GHG-level
    ghg_end : float
        GHG-level in the last period
    emission_by_decisions : ndarray
        emissions at decision time periods??
    emission_per_period : ndarray
        total emission at decision time period??
    emission_to_ghg : ndarray
        GHG levels in decision time period??
    emission_to_bau : float
        constant for converting GHG to emission??
    emit_time : ndarray or list
        time, in years, from now when emissions occurs
    emit_level : ndarray or list
        emission levels in future times `emit_time`

    """
    def __init__(self, ghg_start=400.0, ghg_end=1000.0, emit_time=[0, 30, 60], emit_level=[52.0, 70.0, 81.4]):
        super(DLWBusinessAsUsual, self).__init__(ghg_start, ghg_end)
        self.emit_time = emit_time
        self.emit_level = emit_level

    def emission_by_time(self, time):
        """Returns the BAU emissions at any time

        Parameters
        ----------
        time : int 
            future time period in years

        Returns
        -------
        float
            emission

        """
        if time < self.emit_time[1]:
            emissions = self.emit_level[0] + float(time) / (self.emit_time[1] - self.emit_time[0]) \
                        * (self.emit_level[1] - self.emit_level[0])
        elif time < self.emit_time[2]:
            emissions = self.emit_level[1] + float(time - self.emit_time[1]) / (self.emit_time[2] 
                        - self.emit_time[1]) * (self.emit_level[2] - self.emit_level[1])
        else:
            emissions = self.emit_level[2]
        return emissions

    def bau_emissions_setup(self, tree):
        """Create default business as usual emissions path. The emission rate in each period is 
        assumed to be the average of the emissions at the beginning and at the end of the period.

        Parameters
        ----------
        tree : `TreeModel` object
            provides the tree structure used
            
        """
        num_periods = tree.num_periods
        self.emission_by_decisions = np.zeros(num_periods)
        self.emission_per_period = np.zeros(num_periods)
        self.bau_path = np.zeros(num_periods)
        self.bau_path[0] = self.ghg_start
        self.emission_by_decisions[0] = self.emission_by_time(tree.decision_times[0])
        period_len = tree.decision_times[1:] - tree.decision_times[:-1]

        for n in range(1, num_periods):
            self.emission_by_decisions[n] = self.emission_by_time(tree.decision_times[n])
            self.emission_per_period[n] = period_len[n] * (self.emission_by_decisions[n-1:n].mean())

        #the total increase in ghg level of 600 (from 400 to 1000) in the bau path is allocated over time
        self.emission_to_ghg = (self.ghg_end - self.ghg_start) * self.emission_per_period / self.emission_per_period.sum()
        self.emission_to_bau = self.emission_to_ghg[-1] / self.emission_per_period[-1]
        for n in range(1, num_periods):
            self.bau_path[n] = self.bau_path[n-1] + self.emission_per_period[n]*self.emission_to_bau






from __future__ import division, print_function
import numpy as np
from scipy.optimize import brentq
from storage_tree import BigStorageTree, SmallStorageTree
from optimization import GeneticAlgorithm, GradientSearch
from tools import write_columns_csv, append_to_existing, import_csv


def additional_ghg_emission(m, utility):
	"""Calculate the emission added by every node.

	Parameters
	----------
	m : ndarray or list
		array of mitigation
	utility : `Utility` object
		object of utility class
	
	Returns
	-------
	ndarray
		additional emission in nodes
	
	"""
	additional_emission = np.zeros(len(m))
	cache = set()
	for node in range(utility.tree.num_final_states, len(m)):
		path = utility.tree.get_path(node)
		for i in range(len(path)):
			if path[i] not in cache:
				additional_emission[path[i]] = (1.0 - m[path[i]]) *  utility.damage.bau.emission_to_ghg[i]
				cache.add(path[i])
	return additional_emission

def store_trees(prefix=None, start_year=2015, **kwargs):
	"""Saves values of `BaseStorageTree` objects. The file is saved into the 'data' directory
	in the current working directory. If there is no 'data' directory, one is created. 

	Parameters
	----------
	prefix : str, optional 
		prefix to be added to file_name
	start_year : int, optional
		start year of analysis
	**kwargs 
		arbitrary keyword arguments of `BaseStorageTree` objects

	"""
	if prefix is None:
		prefix = ""
	for name, tree in kwargs.items():
		tree.write_columns(prefix + "trees", name, start_year)

def delta_consumption(m, utility, cons_tree, cost_tree, delta_m):
	"""Calculate the changes in consumption and the mitigation cost component 
	of consumption when increaseing period 0 mitigiation with `delta_m`.

	Parameters
	----------
	m : ndarray or list
		array of mitigation
	utility : `Utility` object
		object of utility class
	cons_tree : `BigStorageTree` object
		consumption storage tree of consumption values
		from optimal mitigation values
	cost_tree : `SmallStorageTree` object
		cost storage tree of cost values from optimal mitigation values
	delta_m : float 
		value to increase period 0 mitigation by
	
	Returns
	-------
	tuple
		(storage tree of changes in consumption, ndarray of costs in first sub periods)

	"""
	m_copy = m.copy()
	m_copy[0] += delta_m

	new_utility_tree, new_cons_tree, new_cost_tree, new_ce_tree = utility.utility(m_copy, return_trees=True)

	for period in new_cons_tree.periods:
		new_cons_tree.tree[period] = (new_cons_tree.tree[period]-cons_tree.tree[period]) / delta_m

	first_period_intervals = new_cons_tree.first_period_intervals
	cost_array = np.zeros((first_period_intervals, 2))
	for i in range(first_period_intervals):
		potential_consumption = (1.0 + utility.cons_growth)**(new_cons_tree.subinterval_len * i)
		cost_array[i, 0] = potential_consumption * cost_tree[0]
		cost_array[i, 1] = (potential_consumption * new_cost_tree[0] - cost_array[i, 0]) / delta_m
	
	return new_cons_tree, cost_array, new_utility_tree[0]

def constraint_first_period(utility, first_node, m_size):
	"""Calculate the changes in consumption, the mitigation cost component of consumption,
	and new mitigation values when constraining the first period mitigation to `first_node`.

	Parameters
	----------
	m : ndarray or list
		array of mitigation
	utility : `Utility` object
		object of utility class
	first_node : float
		value to constrain first period to
	
	Returns
	-------
	tuple
		(new mitigation array, storage tree of changes in consumption, ndarray of costs in first sub periods)

	"""
	fixed_values = np.array([first_node])
	fixed_indicies = np.array([0])
	ga_model = GeneticAlgorithm(pop_amount=150, num_generations=100, cx_prob=0.8, mut_prob=0.5, bound=1.5,
								num_feature=m_size, utility=utility, fixed_values=fixed_values, 
								fixed_indicies=fixed_indicies, print_progress=True)

	gs_model = GradientSearch(var_nums=m_size, utility=utility, accuracy=1e-7,
							  iterations=250, fixed_values=fixed_values, fixed_indicies=fixed_indicies, 
							  print_progress=True)

	final_pop, fitness = ga_model.run()
	sort_pop = final_pop[np.argsort(fitness)][::-1]
	new_m, new_utility = gs_model.run(initial_point_list=sort_pop, topk=1)
	return new_m

def find_ir(m, utility, payment, a=0.0, b=1.0): 
	"""Find the price of a bond that creates equal utility at time 0 as adding `payment` to the value of 
	consumption in the final period. The purpose of this function is to find the interest rate 
	embedded in the `EZUtility` model. 

	Parameters
	----------
	m : ndarray or list
		array of mitigation
	utility : `Utility` object
		object of utility class
	payment : float
		value added to consumption in the final period
	a : float, optional
		initial guess
	b : float, optional
		initial guess - f(b) needs to give different sign than f(a)
	
	Returns
	-------
	tuple
		result of optimization

	Note
	----
	requires the 'scipy' package

	"""

	def min_func(price):
		utility_with_final_payment = utility.adjusted_utility(m, final_cons_eps=payment)
		first_period_eps = payment * price
		utility_with_initial_payment = utility.adjusted_utility(m, first_period_consadj=first_period_eps)
		return utility_with_final_payment - utility_with_initial_payment

	return brentq(min_func, a, b)

def find_term_structure(m, utility, payment, a=0.0, b=1.5): 
	"""Find the price of a bond that creates equal utility at time 0 as adding `payment` to the value of 
	consumption in the final period. The purpose of this function is to find the interest rate 
	embedded in the `EZUtility` model. 

	Parameters
	----------
	m : ndarray or list
		array of mitigation
	utility : `Utility` object
		object of utility class
	payment : float
		value added to consumption in the final period
	a : float, optional
		initial guess
	b : float, optional
		initial guess - f(b) needs to give different sign than f(a)
	
	Returns
	-------
	tuple
		result of optimization

	Note
	----
	requires the 'scipy' package

	"""

	def min_func(price):
		period_cons_eps = np.zeros(int(utility.decision_times[-1]/utility.period_len) + 1)
		period_cons_eps[-2] = payment
		utility_with_payment = utility.adjusted_utility(m, period_cons_eps=period_cons_eps)
		first_period_eps = payment * price
		utility_with_initial_payment = utility.adjusted_utility(m, first_period_consadj=first_period_eps)
		return  utility_with_payment - utility_with_initial_payment

	return brentq(min_func, a, b)

def find_bec(m, utility, constraint_cost, a=-0.1, b=1.5):
	"""Used to find a value for consumption that equalizes utility at time 0 in two different solutions.

	Parameters
	----------
	m : ndarray or list
		array of mitigation
	utility : `Utility` object
		object of utility class
	constraint_cost : float
		utility cost of constraining period 0 to zero
	a : float, optional
		initial guess
	b : float, optional
		initial guess - f(b) needs to give different sign than f(a)
	
	Returns
	-------
	tuple
		result of optimization

	Note
	----
	requires the 'scipy' package

	"""

	def min_func(delta_con):
		base_utility = utility.utility(m)
		new_utility = utility.adjusted_utility(m, first_period_consadj=delta_con)
		print(base_utility, new_utility, constraint_cost)
		return new_utility - base_utility - constraint_cost

	return brentq(min_func, a, b)

def perpetuity_yield(price, start_date, a=0.1, b=10.0):
	"""Find the yield of a perpetuity starting at year `start_date`.

	Parameters
	----------
	price : float
		price of bond ending at `start_date`
	start_date : int
		start year of perpetuity
	a : float, optional
		initial guess
	b : float, optional
		initial guess - f(b) needs to give different sign than f(a)
	
	Returns
	-------
	tuple
		result of optimization

	Note
	----
	requires the 'scipy' package

	"""
	
	def min_func(perp_yield):
		return price - (100. / (perp_yield+100.))**start_date * (perp_yield + 100)/perp_yield

	return brentq(min_func, a, b)


class ClimateOutput(object):
	"""Calculate and save output from the EZ-Climate model.

	Parameters
	----------
	utility : `Utility` object
		object of utility class

	Attributes
	----------
	utility : `Utility` object
		object of utility class
	prices : ndarray
		SCC prices
	ave_mitigations : ndarray
		average mitigations
	ave_emissions : ndarray 
		average emissions
	expected_period_price : ndarray
		expected SCC for the period
 	expected_period_mitigation : ndarray
		expected mitigation for the period
	expected_period_emissions : ndarray
		expected emission for the period

	"""

	def __init__(self, utility):
		self.utility = utility
		self.prices = None
		self.ave_mitigations = None
		self.ave_emissions = None
		self.expected_period_price = None
		self.expected_period_mitigation = None
		self.expected_period_emissions = None
		self.ghg_levels = None

	def calculate_output(self, m):
		"""Calculated values based on optimal mitigation. For every **node** the function calculates and saves
			
			* average mitigation
			* average emission
			* GHG level 
			* SCC 

		as attributes. 

		For every **period** the function also calculates and saves
			
			* expected SCC/price
			* expected mitigation 
			* expected emission 
		
		as attributes.

		Parameters
		----------
		m : ndarray or list
			array of mitigation
	
		"""

		bau = self.utility.damage.bau
		tree = self.utility.tree
		periods = tree.num_periods

		self.prices = np.zeros(len(m))
		self.ave_mitigations = np.zeros(len(m))
		self.ave_emissions = np.zeros(len(m))
		self.expected_period_price = np.zeros(periods)
		self.expected_period_mitigation = np.zeros(periods)
		self.expected_period_emissions = np.zeros(periods)
		additional_emissions = additional_ghg_emission(m, self.utility)
		self.ghg_levels = self.utility.damage.ghg_level(m)

		for period in range(0, periods):
			years = tree.decision_times[period]
			period_years = tree.decision_times[period+1] - tree.decision_times[period]
			nodes = tree.get_nodes_in_period(period)
			num_nodes_period = 1 + nodes[1] - nodes[0]
			period_lens = tree.decision_times[:period+1] 
			
			for node in range(nodes[0], nodes[1]+1):
				path = np.array(tree.get_path(node, period))
				new_m = m[path]
				mean_mitigation = np.dot(new_m, period_lens) / years
				price = self.utility.cost.price(years, m[node], mean_mitigation)
				self.prices[node] = price
				self.ave_mitigations[node] = self.utility.damage.average_mitigation_node(m, node, period)
				self.ave_emissions[node] = additional_emissions[node] / (period_years*bau.emission_to_bau)
			
			probs = tree.get_probs_in_period(period)
			self.expected_period_price[period] = np.dot(self.prices[nodes[0]:nodes[1]+1], probs)
			self.expected_period_mitigation[period] = np.dot(self.ave_mitigations[nodes[0]:nodes[1]+1], probs)
			self.expected_period_emissions[period] = np.dot(self.ave_emissions[nodes[0]:nodes[1]+1], probs)

	def save_output(self, m, prefix=None):
		"""Function to save calculated values in `calculate_output` in the file `prefix` + 'node_period_output' 
		in the 'data' directory in the current working directory. 

		The function also saves the values calculated in the utility function in the file
		`prefix` + 'tree' in the 'data' directory in the current working directory. 

		If there is no 'data' directory, one is created. 

		Parameters
		----------
		m : ndarray or list
			array of mitigation
		prefix : str, optional
			prefix to be added to file_name

		"""
		utility_tree, cons_tree, cost_tree, ce_tree = self.utility.utility(m, return_trees=True)
		
		if prefix is not None:
			prefix += "_" 
		else:
			prefix = ""

		write_columns_csv([m, self.prices, self.ave_mitigations, self.ave_emissions, self.ghg_levels], 
		                   prefix+"node_period_output", ["Node", "Mitigation", "Prices", "Average Mitigation",
		                   "Average Emission", "GHG Level"], [range(len(m))])

		append_to_existing([self.expected_period_price, self.expected_period_mitigation, self.expected_period_emissions],
							prefix+"node_period_output", header=["Period", "Expected Price", "Expected Mitigation",
							"Expected Emission"], index=[range(self.utility.tree.num_periods)], start_char='\n')

		store_trees(prefix=prefix, Utility=utility_tree, Consumption=cons_tree, 
			        Cost=cost_tree, CertainEquivalence=ce_tree)


class RiskDecomposition(object):
	"""Calculate and save analysis of output from the EZ-Climate model.

	Parameters
	----------
	utility : `Utility` object
		object of utility class

	Attributes
	----------
	utility : `Utility` object
		object of utility class
	sdf_tree : `BaseStorageTree` object
		SDF for each node
	expected_damages : ndarray
		expected damages in each period
	risk_premium : ndarray
		risk premium in each period
	expected_sdf : ndarray
		expected SDF in each period
	cross_sdf_damages : ndarray
		cross term between the SDF and damages
	discounted_expected_damages : ndarray
		expected discounted damages for each period
	net_discount_damages : ndarray
		net discount damage, i.e. when cost is also accounted for
	cov_term : ndarray 
		covariance between SDF and damages

	"""

	def __init__(self, utility):
		self.utility = utility
		self.sdf_tree = BigStorageTree(utility.period_len, utility.decision_times)
		self.sdf_tree.set_value(0, np.array([1.0]))

		n = len(self.sdf_tree)
		self.expected_damages = np.zeros(n)
		self.risk_premiums = np.zeros(n)
		self.expected_sdf = np.zeros(n)
		self.cross_sdf_damages = np.zeros(n)
		self.discounted_expected_damages = np.zeros(n)
		self.net_discount_damages = np.zeros(n)
		self.cov_term = np.zeros(n)

		self.expected_sdf[0] = 1.0


	def sensitivity_analysis(self, m):
		"""Calculate sensitivity analysis based on the optimal mitigation. For every sub-period, i.e. the 
		periods given by the utility calculations, the function calculates and saves:
			
			* discount prices
			* net expected damages
			* expected damages
			* discounted expected damages
			* risk premium
			* cross SDF & damages
			* covariance between SDF and damages

		as attributes.

		Parameters
		----------
		m : ndarray or list
			array of mitigation
		utility : `Utility` object
			object of utility class
		prefix : str, optional
			prefix to be added to file_name

		"""

		utility_tree, cons_tree, cost_tree, ce_tree = self.utility.utility(m, return_trees=True)
		cost_sum = 0

		self.delta_cons_tree, self.delta_cost_array, delta_utility = delta_consumption(m, self.utility, cons_tree, cost_tree, 0.01)
		mu_0, mu_1, mu_2 = self.utility.marginal_utility(m, utility_tree, cons_tree, cost_tree, ce_tree)
		sub_len = self.sdf_tree.subinterval_len
		i = 1
		for period in self.sdf_tree.periods[1:]:
			node_period = self.sdf_tree.decision_interval(period)
			period_probs = self.utility.tree.get_probs_in_period(node_period)
			expected_damage = np.dot(self.delta_cons_tree[period], period_probs)
			self.expected_damages[i] = expected_damage
			
			if self.sdf_tree.is_information_period(period-self.sdf_tree.subinterval_len):
				total_probs = period_probs[::2] + period_probs[1::2]
				mu_temp = np.zeros(2*len(mu_1[period-sub_len]))
				mu_temp[::2] = mu_1[period-sub_len]
				mu_temp[1::2] = mu_2[period-sub_len]
				sdf = (np.repeat(total_probs, 2) / period_probs) * (mu_temp/np.repeat(mu_0[period-sub_len], 2))
				period_sdf = np.repeat(self.sdf_tree.tree[period-sub_len],2)*sdf 
			else:
				sdf = mu_1[period-sub_len]/mu_0[period-sub_len]
				period_sdf = self.sdf_tree[period-sub_len]*sdf 

			self.expected_sdf[i] = np.dot(period_sdf, period_probs)
			self.cross_sdf_damages[i] = np.dot(period_sdf, self.delta_cons_tree[period]*period_probs)
			self.cov_term[i] = self.cross_sdf_damages[i] - self.expected_sdf[i]*expected_damage

			self.sdf_tree.set_value(period, period_sdf)

			if i < len(self.delta_cost_array):
				self.net_discount_damages[i] = -(expected_damage + self.delta_cost_array[i, 1]) * self.expected_sdf[i] / self.delta_cons_tree[0]
				cost_sum += -self.delta_cost_array[i, 1] * self.expected_sdf[i] / self.delta_cons_tree[0]
			else:
				self.net_discount_damages[i] = -expected_damage * self.expected_sdf[i] / self.delta_cons_tree[0]

			self.risk_premiums[i] = -self.cov_term[i]/self.delta_cons_tree[0]
			self.discounted_expected_damages[i] = -expected_damage * self.expected_sdf[i] / self.delta_cons_tree[0]
			i += 1

	def save_output(self, m, prefix=None):
		"""Save attributes calculated in `sensitivity_analysis` into the file prefix + `sensitivity_output`
		in the `data` directory in the current working directory.

		Furthermore, the perpetuity yield, the discount factor for the last period is calculated, and SCC,
		expected damage and risk premium for the first period is calculated and saved in into the file
		prefix + `tree` in the `data` directory in the current working directory. If there is no `data` directory, 
		one is created.

		Parameters
		----------
		m : ndarray or list
			array of mitigation
		prefix : str, optional
			prefix to be added to file_name

		"""
		end_price = find_term_structure(m, self.utility, 0.01)
		perp_yield = perpetuity_yield(end_price, self.sdf_tree.periods[-2])

		damage_scale = self.utility.cost.price(0, m[0], 0) / (self.net_discount_damages.sum()+self.risk_premiums.sum())
		scaled_discounted_ed = self.net_discount_damages * damage_scale
		scaled_risk_premiums = self.risk_premiums * damage_scale

		if prefix is not None:
			prefix += "_" 
		else:
			prefix = ""

		write_columns_csv([self.expected_sdf, self.net_discount_damages, self.expected_damages, self.risk_premiums, 
			               self.cross_sdf_damages, self.discounted_expected_damages, self.cov_term, 
			               scaled_discounted_ed, scaled_risk_premiums], prefix + "sensitivity_output",
						   ["Year", "Discount Prices", "Net Expected Damages", "Expected Damages", "Risk Premium",
						    "Cross SDF & Damages", "Discounted Expected Damages", "Cov Term", "Scaled Net Expected Damages",
						    "Scaled Risk Premiums"], [self.sdf_tree.periods.astype(int)+2015]) 

		append_to_existing([[end_price], [perp_yield], [scaled_discounted_ed.sum()], [scaled_risk_premiums.sum()], 
			                [self.utility.cost.price(0, m[0], 0)]], prefix+"sensitivity_output",
			                header=["Zero Bound Price", "Perp Yield", "Expected Damages", "Risk Premium", 
							"SCC"], start_char='\n')
		
		store_trees(prefix=prefix, SDF=self.sdf_tree, DeltaConsumption=self.delta_cons_tree)


class ConstraintAnalysis(object):
	def __init__(self, run_name, utility, const_value, opt_m=None):
		self.run_name = run_name
		self.utility = utility 
		self.cfp_m = constraint_first_period(utility, const_value, utility.tree.num_decision_nodes)
		self.opt_m = opt_m
		if self.opt_m is None:
			self.opt_m = self._get_optimal_m()

		self.con_cost = self._constraint_cost()
		self.delta_u = self._first_period_delta_udiff()

		self.delta_c = self._delta_consumption()
		self.delta_c_billions = self.delta_c * self.utility.cost.cons_per_ton \
								* self.utility.damage.bau.emit_level[0]
		self.delta_emission_gton = self.opt_m[0]*self.utility.damage.bau.emit_level[0]
		self.deadweight = self.delta_c*self.utility.cost.cons_per_ton / self.opt_m[0]

		self.delta_u2 = self._first_period_delta_udiff2()
		self.marginal_benefit = (self.delta_u2 / self.delta_u) * self.utility.cost.cons_per_ton
		self.marginal_cost = self.utility.cost.price(0, self.cfp_m[0], 0)

	def _get_optimal_m(self):
		try:
			header, index, data = import_csv(self.run_name+"_node_period_output")
		except:
			print("No such file for the optimal mitigation..")
		return data[:, 0] 

	def _constraint_cost(self):
		opt_u = self.utility.utility(self.opt_m)
		cfp_u = self.utility.utility(self.cfp_m)
		return opt_u - cfp_u

	def _delta_consumption(self):
		return find_bec(self.cfp_m, self.utility, self.con_cost)

	def _first_period_delta_udiff(self):
		u_given_delta_con = self.utility.adjusted_utility(self.cfp_m, first_period_consadj=0.01)
		cfp_u = self.utility.utility(self.cfp_m)
		return u_given_delta_con - cfp_u

	def _first_period_delta_udiff2(self):
		m = self.cfp_m.copy()
		m[0] += 0.01
		u = self.utility.utility(m)
		cfp_u = self.utility.utility(self.cfp_m)
		return u - cfp_u
		
	def save_output(self, prefix=None):
		if prefix is not None:
			prefix += "_" 
		else:
			prefix = ""

		write_columns_csv([self.con_cost, [self.delta_c], [self.delta_c_billions], [self.delta_emission_gton],
						   [self.deadweight], self.delta_u, self.marginal_benefit, [self.marginal_cost]], 
						   prefix + self.run_name + "_constraint_output",
						  ["Constraint Cost", "Delta Consumption", "Delta Consumption $b", 
						   "Delta Emission Gton", "Deadweight Cost", "Marginal Impact Utility",
						   "Marginal Benefit Emissions Reduction", "Marginal Cost Emission Reduction"])


from __future__ import division
import numpy as np
from abc import ABCMeta, abstractmethod

class BaseStorageTree(object):
	"""Abstract storage class for the EZ-Climate model.

	Parameters
	----------
	decision_times : ndarray or list
		array of years from start where decisions about mitigation levels are done
	
	Attributes
	----------
	decision_times : ndarray 
		array of years from start where decisions about mitigation levels are done
	information_times : ndarray 
		array of years where new information is given to the agent in the model
	periods : ndarray 
		periods in the tree
	tree : dict
		dictionary where keys are `periods` and values are nodes in period

	"""
	__metaclass__ = ABCMeta

	def __init__(self, decision_times):
		self.decision_times = decision_times
		if isinstance(decision_times, list):
			self.decision_times = np.array(decision_times)
		self.information_times = self.decision_times[:-2]
		self.periods = None
		self.tree = None

	def __len__(self):
		return len(self.tree)

	def __getitem__(self, key):
		if isinstance(key, int) or isinstance(key, float):
			return self.tree.__getitem__(key).copy()
		else:
			raise TypeError('Index must be int, not {}'.format(type(key).__name__))

	def _init_tree(self):
		self.tree = dict.fromkeys(self.periods)
		i = 0
		for key in self.periods:
			self.tree[key] = np.zeros(2**i)
			if key in self.information_times:
				i += 1
	@property
	def last(self):
		"""ndarray: last period's array."""
		return self.tree[self.decision_times[-1]]

	@property
	def last_period(self):
		"""int: index of last period."""
		return self.decision_times[-1]

	@property
	def nodes(self):
		"""int: number of nodes in the tree."""
		n = 0
		for array in self.tree.values():
			n += len(array)
		return n

	@abstractmethod
	def get_next_period_array(self, period):
		"""Return the array of the next period from `periods`."""
		pass

	def set_value(self, period, values):
		"""If period is in periods, set the value of element to `values` (ndarray)."""
		if period not in self.periods:
			raise ValueError("Not a valid period")
		if isinstance(values, list):
			values = np.array(values)
		if self.tree[period].shape != values.shape:
			raise ValueError("shapes {} and {} not aligned".format(self.tree[period].shape, values.shape))
		self.tree[period] = values

	def is_decision_period(self, time_period):
		"""Checks if time_period is a decision time for mitigation, where
		time_period is the number of years since start.

		Parameters
		----------
		time_period : int
			time since the start year of the model

		Returns
		-------
		bool
			True if time_period also is a decision time, else False

		"""
		return time_period in self.decision_times

	def is_real_decision_period(self, time_period):
		"""Checks if time_period is a decision time besides the last period, where
		time_period is the number of years since start.
		
		Parameters
		----------
		time_period : int
			time since the start year of the model

		Returns
		-------
		bool
			True if time_period also is a real decision time, else False

		"""
		return time_period in self.decision_times[:-1]

	def is_information_period(self, time_period):
		"""Checks if time_period is a information time for fragility, where
		time_period is the number of years since start.

		Parameters
		----------
		time_period : int
			time since the start year of the model

		Returns
		-------
		bool
			True if time_period also is an information time, else False

		"""
		return time_period in self.information_times

	def write_tree(self, file_name, header, delimiter=";"):
		"""Save values in `tree` as a tree into file  `file_name` in the 
		'data' directory in the current working directory. If there is no 'data' 
		directory, one is created. 
 
		Parameters
		----------
		file_name : str
			name of saved file
		header : str
			first row of file
		delimiter : str, optional
			delimiter in file

		"""
		from tools import find_path
		import csv
		
		real_times = self.decision_times[:-1]
		size = len(self.tree[real_times[-1]])
		output_lst = []
		prev_k = size

		for t in real_times:
			temp_lst = [""]*(size*2)
			k = int(size/len(self.tree[t]))
			temp_lst[k::prev_k] = self.tree[t].tolist()
			output_lst.append(temp_lst)
			prev_k = k

		write_lst = zip(*output_lst)
		d = find_path(file_name)
		with open(d, 'wb') as f:
			writer = csv.writer(f, delimiter=delimiter)
			writer.writerow([header])
			for row in write_lst:
				writer.writerow(row)
	
	def write_columns(self, file_name, header, start_year=2015, delimiter=";"):
		"""Save values in `tree` as columns into file  `file_name` in the 
		'data' directory in the current working directory. If there is no 'data' 
		directory, one is created. 
			
		+------------+------------+-----------+
		|    Year    |    Node 	  |  header   |
		+============+============+===========+
		| start_year |     0	  |   val0    |
		+------------+------------+-----------+
		|     ..     |	   .. 	  |    ..     |
		+------------+------------+-----------+
		
		Parameters
		----------
		file_name : str
			name of saved file
		header : str
			description of values in tree
		start_year : int, optional
			start year of analysis
		delimiter : str, optional
			delimiter in file

		"""
		from tools import write_columns_csv, file_exists
		if file_exists(file_name):
			self.write_columns_existing(file_name, header)
		else:
			real_times = self.decision_times[:-1]
			years = []
			nodes = []
			output_lst = []
			k = 0
			for t in real_times:
				for n in range(len(self.tree[t])):
					years.append(t+start_year)
					nodes.append(k)
					output_lst.append(self.tree[t][n])
					k += 1
			write_columns_csv(lst=[output_lst], file_name=file_name, header=["Year", "Node", header], 
							  index=[years, nodes], delimiter=delimiter)

	def write_columns_existing(self, file_name, header, delimiter=";"):
		"""Save values in `tree` as columns into file  `file_name` in the 
		'data' directory in the current working directory, when `file_name` already exists. 
		If there is no 'data' directory, one is created. 
			
		+------------+------------+-----------------+------------------+
		|    Year    |    Node    |  other_header   |      header      |
		+============+============+=================+==================+
		| start_year |     0      |   other_val0    | 	    val0       |
		+------------+------------+-----------------+------------------+
		|     ..     |     ..     |       ..        |        ..        |
		+------------+------------+-----------------+------------------+

		Parameters
		----------
		file_name : str
			name of saved file
		header : str
			description of values in tree
		start_year : int, optional
			start year of analysis
		delimiter : str, optional
			delimiter in file

		"""
		from tools import write_columns_to_existing
		output_lst = []
		for t in self.decision_times[:-1]:
			output_lst.extend(self.tree[t])
		write_columns_to_existing(lst=output_lst, file_name=file_name, header=header)


class SmallStorageTree(BaseStorageTree):
	"""Storage tree class for the EZ-Climate model. No storage in nodes between 
	periods in `decision_times`.

	Parameters
	----------
	decision_times : ndarray or list
		array of years from start where decisions about mitigation levels are done

	Attributes
	----------
	decision_times : ndarray 
		array of years from start where decisions about mitigation levels are done
	information_times : ndarray 
		array of years where new information is given to the agent in the model
	periods : ndarray 
		periods in the tree
	tree : dict
		dictionary where keys are `periods` and values are nodes in period

	"""
	def __init__(self, decision_times):
		super(SmallStorageTree, self).__init__(decision_times)
		self.periods = self.decision_times
		self._init_tree()

	def get_next_period_array(self, period):
		"""Returns the array of the next decision period.

		Parameters
		----------
		period : int
			period

		Examples
		--------
		>>> sst = SmallStorageTree([0, 15, 45, 85, 185, 285, 385])
		>>> sst.get_next_period_array(0)
		array([0., 0.])
		>>> sst.get_next_period_array(15)
		array([ 0.,  0.,  0.,  0.])

		Raises
		------
		IndexError
			If `period` is not in real decision times

		"""
		if self.is_real_decision_period(period):
			index = self.decision_times[np.where(self.decision_times==period)[0]+1][0]
			return self.tree[index].copy()
		raise IndexError("Given period is not in real decision times")

	def index_below(self, period):
		"""Returns the key of the previous decision period.

		Parameters
		----------
		period : int
			period

		Examples
		--------
		>>> sst = SmallStorageTree([0, 15, 45, 85, 185, 285, 385])
		>>> sst.index_below(15)
		0

		Raises
		------
		IndexError
			If `period` is not in decision times or first element in decision times

		"""
		if period in self.decision_times[1:]:
			period = self.decision_times[np.where(self.decision_times==period)[0]-1]
			return period[0]
		raise IndexError("Period not in decision times or first period")

class BigStorageTree(BaseStorageTree):
	"""Storage tree class for the EZ-Climate model. Storage in nodes between 
	periods in `decision_times`. 

	Parameters
	----------
	subintervals_len : float
		years between periods in tree
	decision_times : ndarray or list
		array of years from start where decisions about mitigation levels are done

	Attributes
	----------
	decision_times : ndarray 
		array of years from start where decisions about mitigation levels are done
	information_times : ndarray 
		array of years where new information is given to the agent in the model
	periods : ndarray 
		periods in the tree
	tree : dict
		dictionary where keys are `periods` and values are nodes in period
	subintervals_len : float
		years between periods in tree

	"""

	def __init__(self, subinterval_len, decision_times):
		super(BigStorageTree, self).__init__(decision_times)
		self.subinterval_len = subinterval_len
		self.periods = np.arange(0, self.decision_times[-1]+self.subinterval_len,
							 self.subinterval_len)
		self._init_tree()

	@property
	def first_period_intervals(self):
		"""ndarray: the number of subintervals in the first period."""
		return int((self.decision_times[1] - self.decision_times[0]) / self.subinterval_len)

	def get_next_period_array(self, period):
		"""Returns the array of the next period.

		Parameters
		----------
		period : int
			period

		Examples
		--------
		>>> bst = BigStorageTree(5.0, [0, 15, 45, 85, 185, 285, 385])
		>>> sst.get_next_period_array(0)
		array([0., 0.])
		>>> sst.get_next_period_array(10)
		array([ 0.,  0., 0., 0.])

		Raises
		------
		IndexError
			If `period` is not a valid period or too large

		"""
		if period + self.subinterval_len <= self.decision_times[-1]:
			return self.tree[period+self.subinterval_len].copy()
		raise IndexError("Period is not a valid period or too large")

	def between_decision_times(self, period):
		"""Check which decision time the period is between and returns
		the index of the lower decision time. 

		Parameters
		----------
		period : int
			period

		Returns
		-------
		int
			index

		Examples
		--------
		>>> bst = BigStorageTree(5, [0, 15, 45, 85, 185, 285, 385])
		>>> bst.between_decision_times(5)
		0
		>>> bst.between_decision_times(15)
		1

		"""
		if period == 0:
			return 0
		for i in range(len(self.information_times)):
			if self.decision_times[i] <= period and period < self.decision_times[i+1]:
				return i
		return i+1

	def decision_interval(self, period):
		"""Check which interval the period is between.
		
		Parameters
		----------
		period : int
			period

		Returns
		-------
		int
			index

		Examples
		--------
		>>> bst = BigStorageTree(5, [0, 15, 45, 85, 185, 285, 385])
		>>> bst.decision_interval(5)
		1
		>>> bst.between_decision_times(15)
		1
		>>> bst.between_decision_times(20)
		2

		"""
		if period == 0:
			return 0
		for i in range(1, len(self.decision_times)):
			if self.decision_times[i-1] < period and period <= self.decision_times[i]:
				return i
		return i




from __future__ import division
import numpy as np

class TreeModel(object):
    """Tree model for the EZ-Climate model. It provides the structure of a non-recombining tree.

    Parameters
    ----------
    decision_times : ndarray or list
        years in the future where decisions will be made
    prob_scale : float, optional
        scaling constant for probabilities

    Attributes
    ----------
    decision_times : ndarray 
        years in the future where decisions will be made
    prob_scale : float
        scaling constant for probabilities
    node_prob : ndarray
        probability of reaching node from period 0
    final_states_prob : ndarray
        last periods `node_prob`

    """

    def __init__(self, decision_times, prob_scale=1.0):
        self.decision_times = decision_times
        if isinstance(self.decision_times, list):
            self.decision_times = np.array(self.decision_times)
        self.prob_scale = prob_scale
        self.node_prob = None
        self.final_states_prob = None
        self._create_probs()

    @property
    def num_periods(self):
        """int: the number of periods in the tree"""
        return len(self.decision_times)-1

    @property
    def num_decision_nodes(self):
        """int: the number of nodes in tree"""
        return (2**self.num_periods) - 1

    @property
    def num_final_states(self):
        """int: the number of nodes in the last period"""
        return 2**(self.num_periods-1)

    def _create_probs(self):
        """Creates the probabilities of every nodes in the tree structure."""
        self.final_states_prob = np.zeros(self.num_final_states)
        self.node_prob = np.zeros(self.num_decision_nodes)
        self.final_states_prob[0] = 1.0
        sum_probs = 1.0
        next_prob = 1.0

        for n in range(1, self.num_final_states):
            next_prob = next_prob * self.prob_scale**(1.0 / n)
            self.final_states_prob[n] = next_prob
        self.final_states_prob /= np.sum(self.final_states_prob)

        self.node_prob[self.num_final_states-1:] = self.final_states_prob
        for period in range(self.num_periods-2, -1, -1): 
            for state in range(0, 2**period):
                pos = self.get_node(period, state)
                self.node_prob[pos] = self.node_prob[2*pos + 1] + self.node_prob[2*pos + 2]

    def get_num_nodes_period(self, period):
        """Returns the number of nodes in the period.

        Parameters
        ----------
        period : int 
            period

        Returns
        -------
        int 
            number of nodes in period

        Examples
        --------
        >>> t = TreeModel([0, 15, 45, 85, 185, 285, 385])
        >>> t.get_num_nodes_period(2)
        4
        >>> t.get_num_nodes_period(5)
        32

        """
        if period >= self.num_periods:
            return 2**(self.num_periods-1)
        return 2**period
    
    def get_nodes_in_period(self, period):
        """Returns the first and last nodes in the period.

        Parameters
        ----------
        period : int 
            period

        Returns
        -------
        int 
            number of nodes in period

        Examples
        --------
        >>> t = TreeModel([0, 15, 45, 85, 185, 285, 385])
        >>> t.get_nodes_in_period(0)
        (0, 0)
        >>> t.get_nodes_in_period(1)
        (1, 2)
        >>> t.get_nodes_in_period(4)
        (15, 30)

        """
        if period >= self.num_periods:
            period = self.num_periods-1
        nodes = self.get_num_nodes_period(period)
        first_node = self.get_node(period, 0)
        return (first_node, first_node+nodes-1)

    def get_node(self, period, state):
        """Returns the node in period and state provided.

        Parameters
        ----------
        period : int 
            period
        state : int
            state of the node

        Returns
        -------
        int 
            node number

        Examples
        --------
        >>> t = TreeModel([0, 15, 45, 85, 185, 285, 385])
        >>> t.get_node(1, 1)
        2
        >>> t.get_node(4, 10)
        25
        >>> t.get_node(4, 20)
        ValueError: No such state in period 4

        Raises
        ------
        ValueError
            If period is too large or if the state is too large
            for the period.

        """
        if period > self.num_periods:
            raise ValueError("Given period is larger than number of periods")
        if state >= 2**period:
            raise ValueError("No such state in period {}".format(period))
        return 2**period + state - 1

    def get_state(self, node, period=None):
        """Returns the state the node represents.

        Parameters
        ----------
        node : int
            the node
        period : int, optional
            the period

        Returns
        -------
        int 
            state

        Examples
        --------
        >>> t = TreeModel([0, 15, 45, 85, 185, 285, 385])
        >>> t.get_state(0)
        0
        >>> t.get_state(4, 2)
        1

        """
        if node >= self.num_decision_nodes:
            return node - self.num_decision_nodes
        if not period:
            period = self.get_period(node)
        return node - (2**period - 1)

    def get_period(self, node):
        """Returns what period the node is in.

        Parameters
        ----------
        node : int
            the node

        Returns
        -------
        int 
            period

        Examples
        --------
        >>> t = TreeModel([0, 15, 45, 85, 185, 285, 385])
        >>> t.get_period(0)
        0
        >>> t.get_period(4)
        2

        """
        if node >= self.num_decision_nodes: 
            return self.num_periods

        for i in range(0, self.num_periods):
            if int((node+1) / 2**i ) == 1:
                return i

    def get_parent_node(self, child):
        """Returns the previous or parent node of the given child node.

        Parameters
        ----------
        child : int
            the child node

        Returns
        -------
        int 
            partent node

        Examples
        --------
        >>> t = TreeModel([0, 15, 45, 85, 185, 285, 385])
        >>> t.get_parent_node(2)
        0
        >>> t.get_parent_node(4)
        1
        >>> t.get_parent_node(10)
        4

        """
        if child == 0:
            return 0
        if child > self.num_decision_nodes:
            return child - self.num_final_states
        if child % 2 == 0:
            return int((child - 2) / 2)
        else:
            return int((child - 1 ) / 2)

    def get_path(self, node, period=None):
        """Returns the unique path taken to come to given node.

        Parameters
        ----------
        node : int
            the node

        Returns
        -------
        ndarray
            path to get to `node`            

        Examples
        --------
        >>> t = TreeModel([0, 15, 45, 85, 185, 285, 385])
        >>> t.get_path(2)
        array([0, 2])
        >>> t.get_parent_node(4)
        array([0, 1, 4])
        >>> t.get_parent_node(62)
        array([ 0,  2,  6, 14, 30, 62])
        
        """
        if period is None:
            period = self.get_period(node)
        path = [node]
        for i in range(0, period):
            parent = self.get_parent_node(path[i])
            path.append(parent)
        path.reverse()
        return np.array(path)

    def get_probs_in_period(self, period):
        """Returns the probabilities to get from period 0 to nodes in period.

        Parameters
        ----------
        period : int
            the period

        Returns
        -------
        ndarray
            probabilities       

        Examples
        --------
        >>> t = TreeModel([0, 15, 45, 85, 185, 285, 385])
        >>> t.get_probs_in_period(2)
        array([ 0.25,  0.25,  0.25,  0.25])
        >>> t.get_probs_in_period(4)
        array([ 0.0625,  0.0625,  0.0625,  0.0625,  0.0625,  0.0625,  0.0625,
                0.0625,  0.0625,  0.0625,  0.0625,  0.0625,  0.0625,  0.0625,
                0.0625,  0.0625])

        """
        first, last = self.get_nodes_in_period(period)
        return self.node_prob[range(first, last+1)]
    
    def reachable_end_states(self, node, period=None, state=None):
        """Returns what future end states can be reached from given node.

        Parameters
        ----------
        node : int
            the node
        period : int, optional
            the period
        state : int, optional
            the state the node is in

        Returns
        -------
        tuple
            (worst end state, best end state)

        Examples
        --------
        >>> t = TreeModel([0, 15, 45, 85, 185, 285, 385])
        >>> t.reachable_end_states(0)
        (0, 31)
        >>> t.reachable_end_states(10)
        (12, 15)
        >>> t.reachable_end_states(32)
        (1, 1)

        """
        if period is None:
            period = self.get_period(node)
        if period >= self.num_periods:
            return (node - self.num_decision_nodes, node - self.num_decision_nodes)
        if state is None:
            state = self.get_state(node, period)

        k = int(self.num_final_states / 2**period)
        return (k*state, k*(state+1)-1)

  




from __future__ import division
import numpy as np
from abc import ABCMeta, abstractmethod
from storage_tree import BigStorageTree

class Cost(object):
	"""Abstract Cost class for the EZ-Climate model."""
	__metaclass__ = ABCMeta

	@abstractmethod
	def cost(self):
		pass

	@abstractmethod
	def price(self):
		pass


class DLWCost(Cost):
	"""Class to evaluate the cost curve for the EZ-Climate model.

	Parameters
	----------
	tree : `TreeModel` object
		tree structure used
	emit_at_0 : float
		initial GHG emission level
	g : float
		intital scale of the cost function
	a : float
		curvature of the cost function
	join_price : float
		price at which the cost curve is extended
	max_price : float
		price at which carbon dioxide can be removed from atmosphere in unlimited scale
	tech_const : float 
		determines the degree of exogenous technological improvement over time. A number 
			of 1.0 implies 1 percent per yer lower cost
	tech_scale : float
		determines the sensitivity of technological change to previous mitigation
	cons_at_0 : float 
		intital consumption. Default $30460bn based on US 2010 values.

	Attributes
	----------
	tree : `TreeModel` object
		tree structure used
	g : float
		intital scale of the cost function
	a : float
		curvature of the cost function
	max_price : float
		price at which carbon dioxide can be removed from atmosphere in unlimited scale
	tech_const : float 
		determines the degree of exogenous technological improvement over time. A number 
			of 1.0 implies 1 percent per yer lower cost
	tech_scale : float
		determines the sensitivity of technological change to previous mitigation
	cons_at_0 : float 
		intital consumption. Default $30460bn based on US 2010 values.
	cbs_level : float
		constant 
	cbs_deriv : float
		constant
	cbs_b : float
		constant 
	cbs_k : float
		constant
	cons_per_ton : float
		constant 
		
	"""

	def __init__(self, tree, emit_at_0, g, a, join_price, max_price,
				tech_const, tech_scale, cons_at_0):
		self.tree = tree
		self.g = g
		self.a = a
		self.max_price = max_price
		self.tech_const = tech_const
		self.tech_scale = tech_scale
		self.cbs_level = (join_price / (g * a))**(1.0 / (a - 1.0))
		self.cbs_deriv = self.cbs_level / (join_price * (a - 1.0))
		self.cbs_b = self.cbs_deriv * (max_price - join_price) / self.cbs_level
		self.cbs_k = self.cbs_level * (max_price - join_price)**self.cbs_b
		self.cons_per_ton = cons_at_0 / emit_at_0

	def cost(self, period, mitigation, ave_mitigation):
		"""Calculates the mitigation cost for the period. For details about the cost function
		see DLW-paper.

		Parameters
		----------
		period : int 
			period in tree for which mitigation cost is calculated
		mitigation : ndarray
			current mitigation values for period
		ave_mitigation : ndarray
			average mitigation up to this period for all nodes in the period

		Returns
		-------
		ndarray 
			cost

		"""		
		years = self.tree.decision_times[period]
		tech_term = (1.0 - ((self.tech_const + self.tech_scale*ave_mitigation) / 100.0))**years
		cbs = self.g * (mitigation**self.a) 
		bool_arr = (mitigation < self.cbs_level).astype(int)
		if np.all(bool_arr):
			c = (cbs * tech_term) / self.cons_per_ton 
		else:
			base_cbs = self.g * self.cbs_level**self.a
			bool_arr2 = (mitigation > self.cbs_level).astype(int)
			extension = ((mitigation-self.cbs_level) * self.max_price 
						- self.cbs_b*mitigation * (self.cbs_k/mitigation)**(1.0/self.cbs_b) / (self.cbs_b-1.0)
						+ self.cbs_b*self.cbs_level * (self.cbs_k/self.cbs_level)**(1.0/self.cbs_b) / (self.cbs_b-1.0))
			
			c = (cbs * bool_arr + (base_cbs + extension)*bool_arr2) * tech_term / self.cons_per_ton
		return c
	
	def price(self, years, mitigation, ave_mitigation):
		"""Inverse of the cost function. Gives emissions price for any given 
		degree of mitigation, average_mitigation, and horizon.

		Parameters
		----------
		years : int y
			years of technological change so far
		mitigation : float 
			mitigation value in node
		ave_mitigation : float
			average mitigation up to this period

		Returns
		-------
		float 
			the price.

		"""
		tech_term = (1.0 - ((self.tech_const + self.tech_scale*ave_mitigation) / 100))**years
		if mitigation < self.cbs_level:
			return self.g * self.a * (mitigation**(self.a-1.0)) * tech_term
		else:
			return (self.max_price - (self.cbs_k/mitigation)**(1.0/self.cbs_b)) * tech_term














from setuptools import setup, find_packages

try:
    from pypandoc import convert
except ImportError:
    import codecs
    read_md = lambda f: codecs.open(f, 'r', 'utf-8').read()
else:
    read_md = lambda f: convert(f, 'rst')

setup(name='ezclimate',
      version='1.2.1b1',
      description='EZ-Climate model',
      long_description=read_md('README.md'),
      classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4'
      ],
      keywords='EZ-climate, social cost of carbon, SCC, climate pricing',
      url='http://github.com/Litterman/EZClimate',
      author='Robert Litterman, Kent Daniel, Gernot Wagner',
      license='MIT',
      packages=find_packages(),
      install_requires=['numpy',],
      include_package_data=False,
      zip_safe=True
      )




from __future__ import division, print_function
from abc import ABCMeta, abstractmethod
import numpy as np
from storage_tree import BigStorageTree, SmallStorageTree

np.seterr(all='ignore')

class EZUtility(object):
	"""Calculation of Epstein-Zin utility for the EZ-Climate model. 

	The Epstein-Zin utility allows for different rates of substitution across time and 
	states. For specification see DLW-paper.

	Parameters
	----------
	tree : `TreeModel` object 
		tree structure used
	damage : `Damage` object
		class that provides damage methods
	cost : `Cost` object
		class that procides cost methods
	period_len : float
		subinterval length
	eis : float, optional
		elasticity of intertemporal substitution
	ra : float, optional
		risk-aversion
	time_pref : float, optional
		pure rate of time preference
	

	Attributes
	----------
	tree : `TreeModel` object 
		tree structure used
	damage : `Damage` object
		class that provides damage methods
	cost : `Cost` object
		class that procides cost methods
	period_len : float
		subinterval length
	decision_times : ndarray 
		years in the future where decisions will be made
	cons_growth : float
		consumption growth 
	growth_term : float
		1 + cons_growth
	r : float
		the parameter rho from the DLW-paper
	a : float
		the parameter alpha from the DLW-paper
	b : float
		the parameter beta from the DLW-paper

	"""

	def __init__(self, tree, damage, cost, period_len, eis=0.9, ra=7.0, time_pref=0.005, 
		         add_penalty_cost=False, max_penalty=0.0, penalty_scale=1.0):
		self.tree = tree
		self.damage = damage
		self.cost = cost
		self.period_len = period_len
		self.decision_times = tree.decision_times
		self.cons_growth = damage.cons_growth
		self.growth_term = 1.0 + self.cons_growth
		self.r = 1.0 - 1.0/eis
		self.a = 1.0 - ra
		self.b = (1.0-time_pref)**period_len
		self.potential_cons = np.ones(self.decision_times.shape) + self.cons_growth
		self.potential_cons = self.potential_cons ** self.decision_times
		self.add_penalty_cost = add_penalty_cost
		self.max_penalty = max_penalty
		self.penalty_scale = penalty_scale

	def _end_period_utility(self, m, utility_tree, cons_tree, cost_tree):
		"""Calculate the terminal utility."""
		period_ave_mitigation = self.damage.average_mitigation(m, self.tree.num_periods)
		period_damage = self.damage.damage_function(m, self.tree.num_periods)
		damage_nodes = self.tree.get_nodes_in_period(self.tree.num_periods)
		
		period_mitigation = m[damage_nodes[0]:damage_nodes[1]+1]
		period_cost = self.cost.cost(self.tree.num_periods, period_mitigation, period_ave_mitigation)

		continuation = (1.0 / (1.0 - self.b*(self.growth_term**self.r)))**(1.0/self.r)

		cost_tree.set_value(cost_tree.last_period, period_cost)
		period_consumption = self.potential_cons[-1] * (1.0 - period_damage)
		period_consumption[period_consumption<=0.0] = 1e-18
		cons_tree.set_value(cons_tree.last_period, period_consumption)
		utility_tree.set_value(utility_tree.last_period, (1.0 - self.b)**(1.0/self.r) * cons_tree.last * continuation)

	def _end_period_marginal_utility(self, mu_tree_0, mu_tree_1, ce_tree, utility_tree, cons_tree):
		"""Calculate the terminal marginal utility."""
		ce_term = utility_tree.last**self.r - (1.0 - self.b)*cons_tree.last**self.r
		ce_tree.set_value(ce_tree.last_period, ce_term)

		mu_0_last = (1.0 - self.b)*(utility_tree[utility_tree.last_period-self.period_len] / cons_tree.last)**(1.0-self.r)
		mu_tree_0.set_value(mu_tree_0.last_period, mu_0_last)
		mu_0 = self._mu_0(cons_tree[cons_tree.last_period-self.period_len], ce_tree[ce_tree.last_period-self.period_len])
		mu_tree_0.set_value(mu_tree_0.last_period-self.period_len, mu_0)

		next_term = self.b * (1.0 - self.b) / (1.0 - self.b * self.growth_term**self.r)
		mu_1 = utility_tree[utility_tree.last_period-self.period_len]**(1-self.r) * next_term * cons_tree.last**(self.r-1.0)
		mu_tree_1.set_value(mu_tree_1.last_period-self.period_len, mu_1)
		

	def _certain_equivalence(self, period, damage_period, utility_tree):
		"""Caclulate certainty equivalence utility. If we are between decision nodes, i.e. no branching,
		then certainty equivalent utility at time period depends only on the utility next period 
		given information known today. Otherwise the certainty equivalent utility is the ability 
		weighted sum of next period utility over the partition reachable from the state.
		"""
		if utility_tree.is_information_period(period):
			damage_nodes = self.tree.get_nodes_in_period(damage_period+1)
			probs = self.tree.node_prob[damage_nodes[0]:damage_nodes[1]+1]
			even_probs = probs[::2]
			odd_probs = probs[1::2]
			even_util = ((utility_tree.get_next_period_array(period)[::2])**self.a) * even_probs
			odd_util = ((utility_tree.get_next_period_array(period)[1::2])**self.a) * odd_probs
			ave_util = (even_util + odd_util) / (even_probs + odd_probs)
			cert_equiv = ave_util**(1.0/self.a)
		else:
			# no branching implies certainty equivalent utility at time period depends only on
			# the utility next period given information known today
			cert_equiv = utility_tree.get_next_period_array(period)
		
		return cert_equiv

	def _utility_generator(self, m, utility_tree, cons_tree, cost_tree, ce_tree, cons_adj=0.0):
		"""Generator for calculating utility for each utility period besides the terminal utility."""
		periods = utility_tree.periods[::-1]

		for period in periods[1:]:
			damage_period = utility_tree.between_decision_times(period)
			cert_equiv = self._certain_equivalence(period, damage_period, utility_tree)

			if utility_tree.is_decision_period(period+self.period_len):
				damage_nodes = self.tree.get_nodes_in_period(damage_period)
				period_mitigation =  m[damage_nodes[0]:damage_nodes[1]+1]
				period_ave_mitigation = self.damage.average_mitigation(m, damage_period)
				period_cost = self.cost.cost(damage_period, period_mitigation, period_ave_mitigation)
				period_damage = self.damage.damage_function(m, damage_period)
				cost_tree.set_value(cost_tree.index_below(period+self.period_len), period_cost)

			period_consumption = self.potential_cons[damage_period] * (1.0 - period_damage) * (1.0 - period_cost)
			period_consumption[period_consumption <= 0.0] = 1e-18

			if not utility_tree.is_decision_period(period):
				next_consumption = cons_tree.get_next_period_array(period)
				segment = period - utility_tree.decision_times[damage_period]
				interval = segment + utility_tree.subinterval_len

				if utility_tree.is_decision_period(period+self.period_len):
					if period < utility_tree.decision_times[-2]:
						next_cost = cost_tree[period+self.period_len]
						next_consumption *= (1.0 - np.repeat(period_cost,2)) / (1.0 - next_cost)
						next_consumption[next_consumption<=0.0] = 1e-18
						
				if period < utility_tree.decision_times[-2]:
					temp_consumption = next_consumption/np.repeat(period_consumption,2)
					period_consumption = np.sign(temp_consumption)*(np.abs(temp_consumption)**(segment/float(interval))) \
					                     * np.repeat(period_consumption,2)
				else:
					temp_consumption = next_consumption/period_consumption
					period_consumption = np.sign(temp_consumption)*(np.abs(temp_consumption)**(segment/float(interval))) \
					                     * period_consumption
				
			
			if period == 0:
				period_consumption += cons_adj

			ce_term = self.b * cert_equiv**self.r
			ce_tree.set_value(period, ce_term)
			cons_tree.set_value(period, period_consumption)
			u = ((1.0-self.b)*period_consumption**self.r + ce_term)**(1.0/self.r)
			yield u, period

	def utility(self, m, return_trees=False):
		"""Calculating utility for the specific mitigation decisions `m`.

		Parameters
		----------
		m : ndarray or list
			array of mitigations
		return_trees : bool
			True if methid should return trees calculated in producing the utility

		Returns
		-------
		ndarray or tuple 
			tuple of `BaseStorageTree` if return_trees else ndarray with utility at period 0

		Examples:
		---------
		Assuming we have declared a EZUtility object as 'ezu' and have a mitigation array 'm'
		
		>>> ezu.utility(m)
		array([ 9.83391921])
		>>> utility_tree, cons_tree, cost_tree, ce_tree = ezu.utility(m, return_trees=True)

		"""
		utility_tree = BigStorageTree(subinterval_len=self.period_len, decision_times=self.decision_times)
		cons_tree = BigStorageTree(subinterval_len=self.period_len, decision_times=self.decision_times)
		ce_tree = BigStorageTree(subinterval_len=self.period_len, decision_times=self.decision_times)
		cost_tree = SmallStorageTree(decision_times=self.decision_times)
	
		self._end_period_utility(m, utility_tree, cons_tree, cost_tree)
		it = self._utility_generator(m, utility_tree, cons_tree, cost_tree, ce_tree)
		for u, period in it:
			utility_tree.set_value(period, u)
			
		if return_trees:
			return utility_tree, cons_tree, cost_tree, ce_tree
		return utility_tree[0]


	def adjusted_utility(self, m, period_cons_eps=None, node_cons_eps=None, final_cons_eps=0.0,
						 first_period_consadj=0.0, return_trees=False):
		"""Calculating adjusted utility for sensitivity analysis. Used e.g. to find zero-coupon bond price.
		Values in parameters are used to adjusted the utility in different ways.

		Parameters
		----------
		m : ndarray 
			array of mitigations
		period_cons_eps : ndarray, optional
			array of increases in consumption per period
		node_cons_eps : `SmallStorageTree`, optional 
			increases in consumption per node
		final_cons_eps : float, optional
			value to increase the final utilities by
		first_period_consadj : float, optional 
			value to increase consumption at period 0 by
		return_trees : bool, optional 
			True if method should return trees calculculated in producing the utility

		Returns
		-------
		ndarray or tuple 
			tuple of `BaseStorageTree` if return_trees else ndarray with utility at period 0

		Examples
		---------
		Assuming we have declared a EZUtility object as 'ezu' and have a mitigation array 'm'
		
		>>> ezu.adjusted_utility(m, final_cons_eps=0.1)
		array([ 9.83424045])
		>>> utility_tree, cons_tree, cost_tree, ce_tree = ezu.adjusted_utility(m, final_cons_eps=0.1, return_trees=True)
		
		>>> arr = np.zeros(int(ezu.decision_times[-1]/ezu.period_len) + 1)
		>>> arr[-1] = 0.1
		>>> ezu.adjusted_utility(m, period_cons_eps=arr)
		array([ 9.83424045])
		
		>>> bst = BigStorageTree(5.0, [0, 15, 45, 85, 185, 285, 385])
		>>> bst.set_value(bst.last_period, np.repeat(0.01, len(bst.last)))
		>>> ezu.adjusted_utility(m, node_cons_eps=bst)
		array([ 9.83391921])

		The last example differs from the rest in that the last values of the `node_cons_eps` will never be
		used. Hence if you want to update the last period consumption, use one of these two methods. 

		>>> ezu.adjusted_utility(m, first_period_consadj=0.01)
		array([ 9.84518772])

		"""
		utility_tree = BigStorageTree(subinterval_len=self.period_len, decision_times=self.decision_times)
		cons_tree = BigStorageTree(subinterval_len=self.period_len, decision_times=self.decision_times)
		ce_tree = BigStorageTree(subinterval_len=self.period_len, decision_times=self.decision_times)
		cost_tree = SmallStorageTree(decision_times=self.decision_times)

		periods = utility_tree.periods[::-1]
		if period_cons_eps is None:
			period_cons_eps = np.zeros(len(periods))
		if node_cons_eps is None:
			node_cons_eps = BigStorageTree(subinterval_len=self.period_len, decision_times=self.decision_times)
		
		self._end_period_utility(m, utility_tree, cons_tree, cost_tree)

		it = self._utility_generator(m, utility_tree, cons_tree, cost_tree, ce_tree, first_period_consadj)
		i = len(utility_tree)-2
		for u, period in it:
			if period == periods[1]:
				mu_0 = (1.0-self.b) * (u/cons_tree[period])**(1.0-self.r)
				next_term = self.b * (1.0-self.b) / (1.0-self.b*self.growth_term**self.r)
				mu_1 = (u**(1.0-self.r)) * next_term * (cons_tree.last**(self.r-1.0))
				u += (final_cons_eps+period_cons_eps[-1]+node_cons_eps.last) * mu_1
				u +=  (period_cons_eps[i]+node_cons_eps.tree[period]) * mu_0
				utility_tree.set_value(period, u)
			else:
				mu_0, m_1, m_2 = self._period_marginal_utility(mu_0, mu_1, m, period, utility_tree, cons_tree, ce_tree)
				u +=  (period_cons_eps[i] + node_cons_eps.tree[period])*mu_0
				utility_tree.set_value(period, u)
			i -= 1

		if return_trees:
			return utility_tree, cons_tree, cost_tree, ce_tree
		return utility_tree.tree[0]

	def _mu_0(self, cons, ce_term):
		"""Marginal utility with respect to consumption function."""
		t1 = (1.0 - self.b)*cons**(self.r-1.0)
		t2 = (ce_term - (self.b-1.0)*cons**self.r)**((1.0/self.r)-1.0)
		return t1 * t2

	def _mu_1(self, cons, prob, cons_1, cons_2, ce_1, ce_2, do_print=False):
		""" marginal utility with respect to consumption next period."""
		t1 = (1.0-self.b) * self.b * prob * cons_1**(self.r-1.0)
		t2 = (ce_1 - (self.b-1.0) * cons_1**self.r )**((self.a/self.r)-1)
		t3 = (prob * (ce_1 - (self.b*(cons_1**self.r)) + cons_1**self.r)**(self.a/self.r) \
			 + (1.0-prob) * (ce_2 - (self.b-1.0) * cons_2**self.r)**(self.a/self.r))**((self.r/self.a)-1.0)
		t4 = prob * (ce_1-self.b * (cons_1**self.r) + cons_1**self.r)**(self.a/self.r) \
			 + (1.0-prob) * (ce_2 - self.b * (cons_2**self.r) + cons_2**self.r)**(self.a/self.r)
		t5 = (self.b * t4**(self.r/self.a) - (self.b-1.0) * cons**self.r )**((1.0/self.r)-1.0)

		return t1 * t2 * t3 * t5 

	def _mu_2(self, cons, prev_cons, ce_term):
		"""Marginal utility with respect to last period consumption."""
		t1 = (1.0-self.b) * self.b * prev_cons**(self.r-1.0)
		t2 = ((1.0 - self.b) * cons**self.r - (self.b - 1.0) * self.b \
		     * prev_cons**self.r + self.b * ce_term)**((1.0/self.r)-1.0)
		return t1 * t2

	def _period_marginal_utility(self, prev_mu_0, prev_mu_1, m, period, utility_tree, cons_tree, ce_tree):
		"""Marginal utility for each node in a period."""
		damage_period = utility_tree.between_decision_times(period)
		mu_0 = self._mu_0(cons_tree[period], ce_tree[period])
		
		prev_ce = ce_tree.get_next_period_array(period)
		prev_cons = cons_tree.get_next_period_array(period)
		if utility_tree.is_information_period(period):
			probs = self.tree.get_probs_in_period(damage_period+1)
			up_prob = np.array([probs[i]/(probs[i]+probs[i+1]) for i in range(0, len(probs), 2)])
			down_prob = 1.0 - up_prob

			up_cons = prev_cons[::2]
			down_cons = prev_cons[1::2]
			up_ce = prev_ce[::2]
			down_ce = prev_ce[1::2]
			
			mu_1 = self._mu_1(cons_tree[period], up_prob, up_cons, down_cons, up_ce, down_ce)
			mu_2 = self._mu_1(cons_tree[period], down_prob, down_cons, up_cons, down_ce, up_ce)
			return mu_0, mu_1, mu_2 
		else:
			mu_1 = self._mu_2(cons_tree[period], prev_cons, prev_ce)
			return mu_0, mu_1, None 

	def marginal_utility(self, m, utility_tree, cons_tree, cost_tree, ce_tree):
		"""Calculating marginal utility for sensitivity analysis, e.g. in the SSC decomposition.

		Parameters
		----------
		m : ndarray
			array of mitigations
		utility_tree : `BigStorageTree` object 
			utility values from using mitigation `m`
		cons_tree : `BigStorageTree` object
			consumption values from using mitigation `m`
		cost_tree : `SmallStorageTree` object
			cost values from using mitigation `m`
		ce_tree : `BigStorageTree` object
			certain equivalence values from using mitigation `m`

		Returns
		-------
		tuple
			marginal utility tree

		Examples
		--------
		Assuming we have declared a EZUtility object as 'ezu' and have a mitigation array 'm'.
		>>> 
		>>> utility_tree, cons_tree, cost_tree, ce_tree = ezu.utility(m, return_trees=True)
		>>> mu_0_tree, mu_1_tree, mu_2_tree = ezu.marginal_utility(m, utility_tree, cons_tree, cost_tree, ce_tree)
		>>> mu_0_tree[0] # value at period 0
		array([ 0.33001256])
		>>> mu_1_tree[0] # value at period 0
		array([ 0.15691619])
		>>> mu_2_tree[0] # value at period 0
		array([ 0.13948175])

		"""
		mu_tree_0 = BigStorageTree(subinterval_len=self.period_len, decision_times=self.decision_times)
		mu_tree_1 = BigStorageTree(subinterval_len=self.period_len, decision_times=self.decision_times)
		mu_tree_2 = SmallStorageTree(decision_times=self.decision_times)
		
		self._end_period_marginal_utility(mu_tree_0, mu_tree_1, ce_tree, utility_tree, cons_tree)
		periods = utility_tree.periods[::-1]

		for period in periods[2:]:
			mu_0, mu_1, mu_2 = self._period_marginal_utility(mu_tree_0.get_next_period_array(period),
				mu_tree_1.get_next_period_array(period), m, period, utility_tree, cons_tree, ce_tree)
			mu_tree_0.set_value(period, mu_0)
			mu_tree_1.set_value(period, mu_1)
			if mu_2 is not None:
				mu_tree_2.set_value(period, mu_2)

		return mu_tree_0, mu_tree_1, mu_tree_2

	def partial_grad(self, m, i, delta=1e-8):
		"""Calculate the ith element of the gradient vector.

		Parameters
		----------
		m : ndarray
			array of mitigations
		i : int
			node to calculate partial grad for

		Returns
		-------
		float
			gradient element

		"""
		m_copy = m.copy()
		m_copy[i] -= delta
		minus_utility = self.utility(m_copy)
		m_copy[i] += 2*delta
		plus_utility = self.utility(m_copy)
		grad = (plus_utility-minus_utility) / (2*delta)
		return grad



from __future__ import division
import numpy as np
import multiprocessing as mp
from tools import _pickle_method, _unpickle_method
from tools import write_columns_csv, append_to_existing
try:
    import copy_reg
except:
    import copyreg as copy_reg
import types

copy_reg.pickle(types.MethodType, _pickle_method, _unpickle_method)

class DamageSimulation(object):
    """Simulation of damages for the EZ-Climate model.

    The damage function simulation is a key input into the pricing engine. Damages are 
    represented in arrays of dimension n x p, where n = num states and p = num periods.
    The arrays are created by Monte Carlo simulation. Each array specifies for each state 
    and time period a damage coefficient. 

    Up to a point, the Monte Carlo follows Pindyck (2012) 'Uncertain Outcomes and Climate Change
    Policy':
    
        * There is a gamma distribution for temperature
        * There is a gamma distribution for economic impact (conditional on temperature)

    However, in addition, this program adds a probability of a tipping point (conditional on temperature).
    This probability is a decreasing function of the parameter `peak_temp`, conditional on a tipping
    point. Damage itself is a decreasing function of the parameter `disaster_tail`.

    Parameters
    ----------
    tree : `TreeModel` object
        tree structure used
    ghg_levels : ndarray or list
        end GHG level for each path
    peak_temp : float
        tipping point parameter 
    disaster_tail : float
        curvature of tipping point
    tip_on : bool
        flag that turns tipping points on or off
    temp_map : int
        mapping from GHG to temperature

            * 0: implies Pindyck displace gamma
            * 1: implies Wagner-Weitzman normal
            * 2: implies Roe-Baker
            * 3: implies user-defined normal 
            * 4: implies user-defined gamma

    temp_dist_params : ndarray or list
        if temp_map is either 3 or 4, user needs to define the distribution parameters
    maxh : float
        time paramter from Pindyck which indicates the time it takes for temp to get half 
        way to its max value for a given level of ghg
    cons_growth : float 
        yearly growth in consumption

    Attributes
    ----------
    tree : `TreeModel` object
        tree structure used
    ghg_levels : ndarray or list
        end GHG level for each path
    peak_temp : float
        tipping point parameter 
    disaster_tail : float
        curvature of tipping point
    tip_on : bool
        flag that turns tipping points on or off
    temp_map : int
        mapping from GHG to temperature
    temp_dist_params : ndarray or list
        if temp_map is either 3 or 4, user needs to define the distribution parameters
    maxh : float
        time paramter from Pindyck which indicates the time it takes for temp to get half 
        way to its max value for a given level of ghg
    cons_growth : float 
        yearly growth in consumption
    d : ndarray
        simulated damages

    """
  
    def __init__(self, tree, ghg_levels, peak_temp, disaster_tail, tip_on, 
                 temp_map, temp_dist_params, maxh, cons_growth):
        self.tree = tree
        self.peak_temp = peak_temp
        self.disaster_tail = disaster_tail
        self.tip_on = tip_on
        self.temp_map = temp_map
        self.dist_params = temp_dist_params
        self.maxh = maxh
        self.cons_growth = cons_growth
        self.ghg_levels = ghg_levels
        self.d = None

    def _write_to_file(self):
        filename = "simulated_damages"
        write_columns_csv(self.d[0].T, filename)
        for arr in self.d[1:]:
            append_to_existing(arr.T, filename, start_char='#')

    def _gamma_array(self, shape, rate, dimension):
        return np.random.gamma(shape, 1.0/rate, dimension)

    def _normal_array(self, mean, stdev, dimension):
        return np.random.normal(mean, stdev, dimension)

    def _uniform_array(self, dimension):
        return np.random.random(dimension)

    def _sort_array(self, array):
        return array[array[:, self.tree.num_periods-1].argsort()]

    def _normal_simulation(self):
        """Draw random samples from normal distribution for mapping GHG to temperature for 
        user-defined distribution parameters.
        """
        assert self.temp_dist_params and len(self.temp_dist_params) == 2, "Normal distribution needs 2 parameters."

        ave, std = temp_dist_params
        n = len(ave)
        temperature = np.array([self._normal_array(ave[i],std[i], self.draws) for i in range(0, n)])
        return np.exp(temperature)

    def _gamma_simulation(self):
        """Draw random samples from gamma distribution for mapping GHG to temperature for 
        user-defined distribution parameters.
        """
        assert self.temp_dist_params and len(self.temp_dist_params) == 3, "Gamma distribution needs 3 parameters."

        k, theta, displace = temp_dist_params
        n = len(k)
        return np.array([self._gamma_array(k[i], theta[i], self.draws) 
                         + displace[i] for i in range(0, n)])

    def _pindyck_simulation(self):
        """Draw random samples for mapping GHG to temperature based on Pindyck. The `pindyck_impact_k` 
        is the shape parameter from Pyndyck damage function, `pindyck_impact_theta` the scale parameter 
        from Pyndyck damage function, and `pindyck_impact_displace` the displacement parameter from Pyndyck
        damage function.
        """
        pindyck_temp_k = [2.81, 4.6134, 6.14]
        pindyck_temp_theta = [1.6667, 1.5974, 1.53139]
        pindyck_temp_displace = [-0.25, -0.5, -1.0]
        return np.array([self._gamma_array(pindyck_temp_k[i], pindyck_temp_theta[i], self.draws) 
                         + pindyck_temp_displace[i] for i in range(0, 3)])

    def _ww_simulation(self):
        """Draw random samples for mapping GHG to temperature based on Wagner-Weitzman."""
        ww_temp_ave = [0.573, 1.148, 1.563]
        ww_temp_stddev = [0.462, 0.441, 0.432]
        temperature = np.array([self._normal_array(ww_temp_ave[i], ww_temp_stddev[i], self.draws) 
                                for i in range(0, 3)])
        return np.exp(temperature)

    def _rb_simulation(self):
        """Draw random samples for mapping GHG to temperature based on Roe-Baker."""        
        rb_fbar = [0.75233, 0.844652, 0.858332]
        rb_sigf = [0.049921, 0.033055, 0.042408]
        rb_theta = [2.304627, 3.333599, 2.356967]
        temperature = np.array([self._normal_array(rb_fbar[i], rb_sigf[i], self.draws) 
                         for i in range(0, 3)])
        return np.maximum(0.0, (1.0 / (1.0 - temperature)) - np.array(rb_theta)[:, np.newaxis])

    def _pindyck_impact_simulation(self):
        """Pindyck gamma distribution mapping temperature into damages."""
        pindyck_impact_k=4.5
        pindyck_impact_theta=21341.0
        pindyck_impact_displace=-0.0000746,
        impact = self._gamma_array(pindyck_impact_k, pindyck_impact_theta, self.draws) + \
                 pindyck_impact_displace 
        return impact

    def _disaster_simulation(self):
        """Simulating disaster random variable, allowing for a tipping point to occur
        with a given probability, leading to a disaster and a `disaster_tail` impact on consumption.
        """
        disaster = self._uniform_array((self.draws, self.tree.num_periods))
        return disaster

    def _disaster_cons_simulation(self):
        """Simulates consumption conditional on disaster, based on the parameter disaster_tail."""
        disaster_cons = self._gamma_array(1.0, self.disaster_tail, self.draws)
        return disaster_cons

    def _interpolation_of_temp(self, temperature): 
        return temperature[:, np.newaxis] * 2.0 * (1.0 - 0.5**(self.tree.decision_times[1:] / self.maxh))
      

    def _economic_impact_of_temp(self, temperature):
        """Economic impact of temperatures, Pindyck [2009]."""
        impact = self._pindyck_impact_simulation()
        term1 = -2.0 * impact[:, np.newaxis] * self.maxh * temperature[:,np.newaxis] / np.log(0.5)
        term2 = (self.cons_growth - 2.0 * impact[:, np.newaxis] \
                * temperature[:, np.newaxis]) * self.tree.decision_times[1:]
        term3 = (2.0 * impact[:, np.newaxis] * self.maxh \
                * temperature[:, np.newaxis] * 0.5**(self.tree.decision_times[1:] / self.maxh)) / np.log(0.5)
        return np.exp(term1 + term2 + term3)

    def _tipping_point_update(self, tmp, consump, peak_temp_interval=30.0):
        """Determine whether a tipping point has occurred, if so reduce consumption for 
        all periods after this date.
        """
        draws = tmp.shape[0]
        disaster = self._disaster_simulation()
        disaster_cons = self._disaster_cons_simulation()
        period_lengths = self.tree.decision_times[1:] - self.tree.decision_times[:-1]
        
        tmp_scale = np.maximum(self.peak_temp, tmp)
        ave_prob_of_survival = 1.0 - np.square(tmp / tmp_scale) 
        prob_of_survival = ave_prob_of_survival**(period_lengths / peak_temp_interval)
        # this part may be done better, this takes a long time to loop over
        res = prob_of_survival < disaster
        rows, cols = np.nonzero(res)
        row, count = np.unique(rows, return_counts=True)
        first_occurance = zip(row, cols[np.insert(count.cumsum()[:-1],0,0)])
        for pos in first_occurance:
            consump[pos[0], pos[1]:] *= np.exp(-disaster_cons[pos[0]])
        return consump

    def _run_path(self, temperature):
        """Calculate the distribution of damage for specific GHG-path. Implementation of 
        the temperature and economic impacts from Pindyck [2012] page 6.
        """
        d = np.zeros((self.tree.num_final_states, self.tree.num_periods))
        tmp = self._interpolation_of_temp(temperature)
        consump = self._economic_impact_of_temp(temperature)
        peak_cons = np.exp(self.cons_growth*self.tree.decision_times[1:])
            
        # adding tipping points
        if self.tip_on:
            consump = self._tipping_point_update(tmp, consump)
                
        # sort based on outcome of simulation
        consump = self._sort_array(consump)
        damage = 1.0 - (consump / peak_cons)
        weights = self.tree.final_states_prob*(self.draws)
        weights = (weights.cumsum()).astype(int)
    
        d[0,] = damage[:weights[0], :].mean(axis=0)
        for n in range(1, self.tree.num_final_states):
            d[n,] = np.maximum(0.0, damage[weights[n-1]:weights[n], :].mean(axis=0))
        return d

    def simulate(self, draws, write_to_file=True):
        """Create damage function values in 'p-period' version of the Summers - Zeckhauser model.

        Parameters
        ----------
        draws : int
            number of samples drawn in Monte Carlo simulation.
        write_to_file : bool, optional
            wheter to save simulated values 
       
        Returns
        -------
        ndarray 
            3D-array of simulated damages 

        Raises
        ------
        ValueError
            If temp_map is not in the interval 0-4.         

        Note
        ----
        Uses the :mod:`~multiprocessing` package.

        """
        dnum = len(self.ghg_levels)
        self.draws = draws 
        self.peak_cons = np.exp(self.cons_growth*self.tree.decision_times[1:])

        if self.temp_map == 0:
            temperature = self._pindyck_simulation()
        elif self.temp_map == 1:
            temperature = self._ww_simulation()
        elif self.temp_map == 2:
            temperature = self._rb_simulation()
        elif self.temp_map == 3:
            temperature = self._normal_simulation()
        elif self.temp_map == 4:
            temperature = self._gamma_simulation()
        else:
            raise ValueError("temp_map not in interval 0-4")

        pool = mp.Pool(processes=dnum)
        self.d = np.array(pool.map(self._run_path, temperature))

        if write_to_file:
            self._write_to_file()
        return self.d


    














